"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_sym_db = _symbol_database.Default()
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dbdk/v2/requests/offload.proto\x12\x06bdk.v2"J\n\x07Offload\x12&\n\x03aws\x18\x01 \x01(\x0b2\x12.bdk.v2.AWSOffloadH\x00R\x03awsB\x17\n\x15offload_discriminator"\xf9\x01\n\nAWSOffload\x12\x1d\n\naccess_key\x18\x01 \x01(\tR\taccessKey\x12\x1d\n\nsecret_key\x18\x02 \x01(\tR\tsecretKey\x12#\n\rsession_token\x18\x03 \x01(\tR\x0csessionToken\x12\x16\n\x06region\x18\x04 \x01(\tR\x06region\x12\x16\n\x06bucket\x18\x05 \x01(\tR\x06bucket\x12\x1f\n\x0bfolder_name\x18\x06 \x01(\tR\nfolderName\x12&\n\x0cendpoint_url\x18\x07 \x01(\tH\x00R\x0bendpointUrl\x88\x01\x01B\x0f\n\r_endpoint_urlB\x1fZ\x1dgithub.com/kognitos/bdk-protob\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'bdk.v2.requests.offload_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals['DESCRIPTOR']._options = None
    _globals['DESCRIPTOR']._serialized_options = b'Z\x1dgithub.com/kognitos/bdk-proto'
    _globals['_OFFLOAD']._serialized_start = 41
    _globals['_OFFLOAD']._serialized_end = 115
    _globals['_AWSOFFLOAD']._serialized_start = 118
    _globals['_AWSOFFLOAD']._serialized_end = 367


from typing import Tuple

import google.protobuf.timestamp_pb2
from google.protobuf.struct_pb2 import \
    NULL_VALUE as PROTO_NULL_VALUE  # pylint: disable=no-name-in-module

# pylint: disable=no-name-in-module
from ..proto.bdk.v1.types.date_pb2 import Date as ProtoDate
from ..proto.bdk.v1.types.noun_phrase_pb2 import NounPhrase as ProtoNounPhrase
from ..proto.bdk.v1.types.time_pb2 import Time as ProtoTime
from ..proto.bdk.v1.types.value_pb2 import Column as ProtoColumn
from ..proto.bdk.v1.types.value_pb2 import \
    DictionaryValue as ProtoDictionaryValue
from ..proto.bdk.v1.types.value_pb2 import \
    DictionaryValueField as ProtoDictionaryValueField
from ..proto.bdk.v1.types.value_pb2 import File as ProtoFile
from ..proto.bdk.v1.types.value_pb2 import FileValue as ProtoFileValue
from ..proto.bdk.v1.types.value_pb2 import ListValue as ProtoListValue
from ..proto.bdk.v1.types.value_pb2 import OpaqueValue as ProtoOpaqueValue
from ..proto.bdk.v1.types.value_pb2 import \
    SensitiveValue as ProtoSensitiveValue
from ..proto.bdk.v1.types.value_pb2 import Table as ProtoTable
from ..proto.bdk.v1.types.value_pb2 import TableValue as ProtoTableValue
from ..proto.bdk.v1.types.value_pb2 import Value as ProtoValue
from ..value import (BooleanValue, ConceptualValue, DatetimeValue, DateValue,
                     DictionaryValue, File, ListValue, NullValue, NumberValue,
                     OpaqueValue, RemoteFile, RemoteTable, SensitiveValue,
                     Table, TextValue, TimeValue, Value)


def _micros_to_nanos(micros: int) -> int:
    return micros * 1000


def _timestamp_to_seconds_and_nanos(datetime: DatetimeValue) -> Tuple[int, int]:
    return int(datetime.timestamp()), _micros_to_nanos(datetime.microsecond)


def map_value(value: Value) -> ProtoValue:
    """Maps a value from bdk-runtime-client domain to proto"""

    if isinstance(value, ConceptualValue):
        return ProtoValue(
            conceptual_value=ProtoNounPhrase(head=value.head, modifiers=value.modifiers)
        )

    if isinstance(value, NumberValue):
        return ProtoValue(number_value=value)

    if isinstance(value, NullValue):
        return ProtoValue(null_value=PROTO_NULL_VALUE)  # type: ignore

    if isinstance(value, TextValue):
        return ProtoValue(text_value=value)

    if isinstance(value, BooleanValue):
        return ProtoValue(boolean_value=value)

    if isinstance(value, OpaqueValue):
        return ProtoValue(
            opaque_value=ProtoOpaqueValue(
                content=value.value,
                is_a=[
                    ProtoNounPhrase(head=is_a.head, modifiers=is_a.modifiers)
                    for is_a in value.is_a
                ],
            )
        )

    if isinstance(value, bytes):
        return ProtoValue(opaque_value=ProtoOpaqueValue(content=value))

    if isinstance(value, DatetimeValue):
        seconds, nanos = _timestamp_to_seconds_and_nanos(value)
        return ProtoValue(
            datetime_value=google.protobuf.timestamp_pb2.Timestamp(  # pylint: disable=no-member
                seconds=seconds, nanos=nanos
            )
        )

    if isinstance(value, DateValue):
        return ProtoValue(
            date_value=ProtoDate(year=value.year, month=value.month, day=value.day)
        )

    if isinstance(value, TimeValue):
        return ProtoValue(
            time_value=ProtoTime(
                hours=value.hour,
                minutes=value.minute,
                seconds=value.second,
                nanos=_micros_to_nanos(value.microsecond),
            )
        )

    if isinstance(value, ListValue):
        return ProtoValue(
            list_value=ProtoListValue(values=[map_value(v) for v in value])
        )

    if isinstance(value, DictionaryValue):
        return ProtoValue(
            dictionary_value=ProtoDictionaryValue(
                fields=[
                    ProtoDictionaryValueField(key=k, value=map_value(v))
                    for k, v in value.value.items()
                ],
                is_a=[
                    ProtoNounPhrase(head=is_a.head, modifiers=is_a.modifiers)
                    for is_a in value.is_a
                ],
            )
        )

    if isinstance(value, dict):
        return ProtoValue(
            dictionary_value=ProtoDictionaryValue(
                fields=[
                    ProtoDictionaryValueField(key=k, value=map_value(v))
                    for k, v in value.items()
                ]
            )
        )

    if isinstance(value, RemoteFile):
        return ProtoValue(file_value=ProtoFileValue(remote=value.url))

    if isinstance(value, File):
        return ProtoValue(
            file_value=ProtoFileValue(
                inline=ProtoFile(file_name=value.filename, content=value.content)
            )
        )

    if isinstance(value, Table):
        return ProtoValue(
            table_value=ProtoTableValue(
                inline=ProtoTable(
                    columns=[
                        ProtoColumn(
                            name=c.name, values=[map_value(val) for val in c.values]
                        )
                        for c in value.columns
                    ]
                )
            )
        )

    if isinstance(value, RemoteTable):
        return ProtoValue(table_value=ProtoTableValue(remote=value.url))

    if isinstance(value, SensitiveValue):
        return ProtoValue(
            sensitive_value=ProtoSensitiveValue(value=map_value(value.value))
        )

    raise NotImplementedError(f"Can't map value '{value}' of type '{type(value)}'")



from bdk.v2.types import book_descriptor_pb2 as _book_descriptor_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union
DESCRIPTOR: _descriptor.FileDescriptor

class RetrieveBooksResponse(_message.Message):
    __slots__ = ('books', 'next_page_token')
    BOOKS_FIELD_NUMBER: _ClassVar[int]
    NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    books: _containers.RepeatedCompositeFieldContainer[_book_descriptor_pb2.BookDescriptor]
    next_page_token: str

    def __init__(self, books: _Optional[_Iterable[_Union[_book_descriptor_pb2.BookDescriptor, _Mapping]]]=..., next_page_token: _Optional[str]=...) -> None:
        ...


"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_sym_db = _symbol_database.Default()
from ....bdk.v1.types import book_descriptor_pb2 as bdk_dot_v1_dot_types_dot_book__descriptor__pb2
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$bdk/v1/responses/retrieve_book.proto\x12\x06bdk.v1\x1a"bdk/v1/types/book_descriptor.proto"B\n\x14RetrieveBookResponse\x12*\n\x04book\x18\x01 \x01(\x0b2\x16.bdk.v1.BookDescriptorR\x04bookB\x1fZ\x1dgithub.com/kognitos/bdk-protob\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'bdk.v1.responses.retrieve_book_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals['DESCRIPTOR']._options = None
    _globals['DESCRIPTOR']._serialized_options = b'Z\x1dgithub.com/kognitos/bdk-proto'
    _globals['_RETRIEVEBOOKRESPONSE']._serialized_start = 84
    _globals['_RETRIEVEBOOKRESPONSE']._serialized_end = 150


"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_sym_db = _symbol_database.Default()
from ....bdk.v2.requests import authentication_pb2 as bdk_dot_v2_dot_requests_dot_authentication__pb2
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n)bdk/v2/requests/discover_procedures.proto\x12\x06bdk.v2\x1a$bdk/v2/requests/authentication.proto"\x9d\x01\n\x19DiscoverProceduresRequest\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x18\n\x07version\x18\x02 \x01(\tR\x07version\x12>\n\x0eauthentication\x18\x03 \x01(\x0b2\x16.bdk.v2.AuthenticationR\x0eauthentication\x12\x12\n\x04what\x18\x04 \x03(\tR\x04whatB\x1fZ\x1dgithub.com/kognitos/bdk-protob\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'bdk.v2.requests.discover_procedures_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals['DESCRIPTOR']._options = None
    _globals['DESCRIPTOR']._serialized_options = b'Z\x1dgithub.com/kognitos/bdk-proto'
    _globals['_DISCOVERPROCEDURESREQUEST']._serialized_start = 92
    _globals['_DISCOVERPROCEDURESREQUEST']._serialized_end = 249


"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_sym_db = _symbol_database.Default()
from ....bdk.v1.types import book_custom_authentication_descriptor_pb2 as bdk_dot_v1_dot_types_dot_book__custom__authentication__descriptor__pb2
from ....bdk.v1.types import book_oauh_authentication_descriptor_pb2 as bdk_dot_v1_dot_types_dot_book__oauh__authentication__descriptor__pb2
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n1bdk/v1/types/book_authentication_descriptor.proto\x12\x06bdk.v1\x1a8bdk/v1/types/book_custom_authentication_descriptor.proto\x1a6bdk/v1/types/book_oauh_authentication_descriptor.proto"\xc7\x01\n\x1cBookAuthenticationDescriptor\x12D\n\x06custom\x18\x01 \x01(\x0b2*.bdk.v1.BookCustomAuthenticationDescriptorH\x00R\x06custom\x12A\n\x05oauth\x18\x02 \x01(\x0b2).bdk.v1.BookOAuthAuthenticationDescriptorH\x00R\x05oauthB\x1e\n\x1cauthentication_discriminatorB\x1fZ\x1dgithub.com/kognitos/bdk-protob\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'bdk.v1.types.book_authentication_descriptor_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals['DESCRIPTOR']._options = None
    _globals['DESCRIPTOR']._serialized_options = b'Z\x1dgithub.com/kognitos/bdk-proto'
    _globals['_BOOKAUTHENTICATIONDESCRIPTOR']._serialized_start = 176
    _globals['_BOOKAUTHENTICATIONDESCRIPTOR']._serialized_end = 375


from bdk.v2.types import book_descriptor_pb2 as _book_descriptor_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union
DESCRIPTOR: _descriptor.FileDescriptor

class RetrieveBookResponse(_message.Message):
    __slots__ = ('book',)
    BOOK_FIELD_NUMBER: _ClassVar[int]
    book: _book_descriptor_pb2.BookDescriptor

    def __init__(self, book: _Optional[_Union[_book_descriptor_pb2.BookDescriptor, _Mapping]]=...) -> None:
        ...


from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional
DESCRIPTOR: _descriptor.FileDescriptor

class ExampleDescriptor(_message.Message):
    __slots__ = ('description', 'snippet')
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    SNIPPET_FIELD_NUMBER: _ClassVar[int]
    description: str
    snippet: str

    def __init__(self, description: _Optional[str]=..., snippet: _Optional[str]=...) -> None:
        ...


from kognitos.bdk.runtime.client.credential_descriptor import (
    CredentialDescriptor, CredentialType)


def map_credential_descriptor(data) -> CredentialDescriptor:
    return CredentialDescriptor(
        id=data.id,
        type=CredentialType(data.type),
        label=data.label,
        visible=data.visible,
        description=data.description,
    )



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_sym_db = _symbol_database.Default()
from ....bdk.v2.types import noun_phrase_pb2 as bdk_dot_v2_dot_types_dot_noun__phrase__pb2
from ....bdk.v2.types import value_pb2 as bdk_dot_v2_dot_types_dot_value__pb2
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dbdk/v2/types/expression.proto\x12\x06bdk.v2\x1a\x1ebdk/v2/types/noun_phrase.proto\x1a\x18bdk/v2/types/value.proto"\xb4\x02\n\nExpression\x12G\n\x11binary_expression\x18\x01 \x01(\x0b2\x18.bdk.v2.BinaryExpressionH\x00R\x10binaryExpression\x12D\n\x10unary_expression\x18\x02 \x01(\x0b2\x17.bdk.v2.UnaryExpressionH\x00R\x0funaryExpression\x12:\n\x10value_expression\x18\x03 \x01(\x0b2\r.bdk.v2.ValueH\x00R\x0fvalueExpression\x12M\n\x17noun_phrases_expression\x18\x04 \x01(\x0b2\x13.bdk.v2.NounPhrasesH\x00R\x15nounPhrasesExpressionB\x0c\n\nexpression"\x98\x01\n\x10BinaryExpression\x122\n\x08operator\x18\x01 \x01(\x0e2\x16.bdk.v2.BinaryOperatorR\x08operator\x12&\n\x04left\x18\x02 \x01(\x0b2\x12.bdk.v2.ExpressionR\x04left\x12(\n\x05right\x18\x03 \x01(\x0b2\x12.bdk.v2.ExpressionR\x05right"n\n\x0fUnaryExpression\x121\n\x08operator\x18\x01 \x01(\x0e2\x15.bdk.v2.UnaryOperatorR\x08operator\x12(\n\x05inner\x18\x02 \x01(\x0b2\x12.bdk.v2.ExpressionR\x05inner*\xa5\x02\n\x0eBinaryOperator\x12\x15\n\x11BinaryOperatorAnd\x10\x00\x12\x14\n\x10BinaryOperatorOr\x10\x01\x12\x18\n\x14BinaryOperatorEquals\x10\x02\x12\x1b\n\x17BinaryOperatorNotEquals\x10\x03\x12\x14\n\x10BinaryOperatorIn\x10\x04\x12\x15\n\x11BinaryOperatorHas\x10\x05\x12\x1a\n\x16BinaryOperatorLessThan\x10\x06\x12\x1d\n\x19BinaryOperatorGreaterThan\x10\x07\x12!\n\x1dBinaryOperatorLessThanOrEqual\x10\x08\x12$\n BinaryOperatorGreaterThanOrEqual\x10\t*%\n\rUnaryOperator\x12\x14\n\x10UnaryOperatorNot\x10\x00B\x1fZ\x1dgithub.com/kognitos/bdk-protob\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'bdk.v2.types.expression_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals['DESCRIPTOR']._options = None
    _globals['DESCRIPTOR']._serialized_options = b'Z\x1dgithub.com/kognitos/bdk-proto'
    _globals['_BINARYOPERATOR']._serialized_start = 678
    _globals['_BINARYOPERATOR']._serialized_end = 971
    _globals['_UNARYOPERATOR']._serialized_start = 973
    _globals['_UNARYOPERATOR']._serialized_end = 1010
    _globals['_EXPRESSION']._serialized_start = 100
    _globals['_EXPRESSION']._serialized_end = 408
    _globals['_BINARYEXPRESSION']._serialized_start = 411
    _globals['_BINARYEXPRESSION']._serialized_end = 563
    _globals['_UNARYEXPRESSION']._serialized_start = 565
    _globals['_UNARYEXPRESSION']._serialized_end = 675


from bdk.v1.responses import discover_procedures_pb2 as _discover_procedures_pb2
from bdk.v1.responses import environment_information_pb2 as _environment_information_pb2
from bdk.v1.responses import error_pb2 as _error_pb2
from bdk.v1.responses import invoke_procedure_pb2 as _invoke_procedure_pb2
from bdk.v1.responses import promise_pb2 as _promise_pb2
from bdk.v1.responses import question_pb2 as _question_pb2
from bdk.v1.responses import retrieve_book_pb2 as _retrieve_book_pb2
from bdk.v1.responses import retrieve_book_procedures_pb2 as _retrieve_book_procedures_pb2
from bdk.v1.responses import retrieve_books_pb2 as _retrieve_books_pb2
from bdk.v1.responses import retrieve_discoverables_pb2 as _retrieve_discoverables_pb2
from bdk.v1.responses import retrieve_tags_pb2 as _retrieve_tags_pb2
from bdk.v1.responses import test_connection_pb2 as _test_connection_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union
DESCRIPTOR: _descriptor.FileDescriptor

class Response(_message.Message):
    __slots__ = ('error', 'environment_information', 'retrieved_books', 'retrieved_book', 'retrieved_procedures', 'tested_connection', 'invoked_procedure', 'discovered_procedures', 'question', 'retrieved_discoverables', 'retrieved_tags', 'promise')
    ERROR_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_INFORMATION_FIELD_NUMBER: _ClassVar[int]
    RETRIEVED_BOOKS_FIELD_NUMBER: _ClassVar[int]
    RETRIEVED_BOOK_FIELD_NUMBER: _ClassVar[int]
    RETRIEVED_PROCEDURES_FIELD_NUMBER: _ClassVar[int]
    TESTED_CONNECTION_FIELD_NUMBER: _ClassVar[int]
    INVOKED_PROCEDURE_FIELD_NUMBER: _ClassVar[int]
    DISCOVERED_PROCEDURES_FIELD_NUMBER: _ClassVar[int]
    QUESTION_FIELD_NUMBER: _ClassVar[int]
    RETRIEVED_DISCOVERABLES_FIELD_NUMBER: _ClassVar[int]
    RETRIEVED_TAGS_FIELD_NUMBER: _ClassVar[int]
    PROMISE_FIELD_NUMBER: _ClassVar[int]
    error: _error_pb2.Error
    environment_information: _environment_information_pb2.EnvironmentInformationResponse
    retrieved_books: _retrieve_books_pb2.RetrieveBooksResponse
    retrieved_book: _retrieve_book_pb2.RetrieveBookResponse
    retrieved_procedures: _retrieve_book_procedures_pb2.RetrieveBookProceduresResponse
    tested_connection: _test_connection_pb2.TestConnectionResponse
    invoked_procedure: _invoke_procedure_pb2.InvokeProcedureResponse
    discovered_procedures: _discover_procedures_pb2.DiscoverProceduresResponse
    question: _question_pb2.QuestionResponse
    retrieved_discoverables: _retrieve_discoverables_pb2.RetrieveDiscoverablesResponse
    retrieved_tags: _retrieve_tags_pb2.RetrieveTagsResponse
    promise: _promise_pb2.PromiseResponse

    def __init__(self, error: _Optional[_Union[_error_pb2.Error, _Mapping]]=..., environment_information: _Optional[_Union[_environment_information_pb2.EnvironmentInformationResponse, _Mapping]]=..., retrieved_books: _Optional[_Union[_retrieve_books_pb2.RetrieveBooksResponse, _Mapping]]=..., retrieved_book: _Optional[_Union[_retrieve_book_pb2.RetrieveBookResponse, _Mapping]]=..., retrieved_procedures: _Optional[_Union[_retrieve_book_procedures_pb2.RetrieveBookProceduresResponse, _Mapping]]=..., tested_connection: _Optional[_Union[_test_connection_pb2.TestConnectionResponse, _Mapping]]=..., invoked_procedure: _Optional[_Union[_invoke_procedure_pb2.InvokeProcedureResponse, _Mapping]]=..., discovered_procedures: _Optional[_Union[_discover_procedures_pb2.DiscoverProceduresResponse, _Mapping]]=..., question: _Optional[_Union[_question_pb2.QuestionResponse, _Mapping]]=..., retrieved_discoverables: _Optional[_Union[_retrieve_discoverables_pb2.RetrieveDiscoverablesResponse, _Mapping]]=..., retrieved_tags: _Optional[_Union[_retrieve_tags_pb2.RetrieveTagsResponse, _Mapping]]=..., promise: _Optional[_Union[_promise_pb2.PromiseResponse, _Mapping]]=...) -> None:
        ...


from bdk.v1.types import noun_phrase_pb2 as _noun_phrase_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union
DESCRIPTOR: _descriptor.FileDescriptor

class ConceptScalarType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ConceptScalarTypeConceptual: _ClassVar[ConceptScalarType]
    ConceptScalarTypeText: _ClassVar[ConceptScalarType]
    ConceptScalarTypeNumber: _ClassVar[ConceptScalarType]
    ConceptScalarTypeBoolean: _ClassVar[ConceptScalarType]
    ConceptScalarTypeDatetime: _ClassVar[ConceptScalarType]
    ConceptScalarTypeDate: _ClassVar[ConceptScalarType]
    ConceptScalarTypeTime: _ClassVar[ConceptScalarType]
    ConceptScalarTypeFile: _ClassVar[ConceptScalarType]
    ConceptScalarTypeUUID: _ClassVar[ConceptScalarType]
ConceptScalarTypeConceptual: ConceptScalarType
ConceptScalarTypeText: ConceptScalarType
ConceptScalarTypeNumber: ConceptScalarType
ConceptScalarTypeBoolean: ConceptScalarType
ConceptScalarTypeDatetime: ConceptScalarType
ConceptScalarTypeDate: ConceptScalarType
ConceptScalarTypeTime: ConceptScalarType
ConceptScalarTypeFile: ConceptScalarType
ConceptScalarTypeUUID: ConceptScalarType

class ConceptOptionalType(_message.Message):
    __slots__ = ('inner',)
    INNER_FIELD_NUMBER: _ClassVar[int]
    inner: ConceptType

    def __init__(self, inner: _Optional[_Union[ConceptType, _Mapping]]=...) -> None:
        ...

class ConceptSensitiveType(_message.Message):
    __slots__ = ('inner',)
    INNER_FIELD_NUMBER: _ClassVar[int]
    inner: ConceptType

    def __init__(self, inner: _Optional[_Union[ConceptType, _Mapping]]=...) -> None:
        ...

class ConceptListType(_message.Message):
    __slots__ = ('inner',)
    INNER_FIELD_NUMBER: _ClassVar[int]
    inner: ConceptType

    def __init__(self, inner: _Optional[_Union[ConceptType, _Mapping]]=...) -> None:
        ...

class ConceptUnionType(_message.Message):
    __slots__ = ('inners',)
    INNERS_FIELD_NUMBER: _ClassVar[int]
    inners: _containers.RepeatedCompositeFieldContainer[ConceptType]

    def __init__(self, inners: _Optional[_Iterable[_Union[ConceptType, _Mapping]]]=...) -> None:
        ...

class ConceptDictionaryType(_message.Message):
    __slots__ = ('is_a', 'fields', 'description')
    IS_A_FIELD_NUMBER: _ClassVar[int]
    FIELDS_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    is_a: _containers.RepeatedCompositeFieldContainer[_noun_phrase_pb2.NounPhrase]
    fields: _containers.RepeatedCompositeFieldContainer[ConceptDictionaryTypeField]
    description: str

    def __init__(self, is_a: _Optional[_Iterable[_Union[_noun_phrase_pb2.NounPhrase, _Mapping]]]=..., fields: _Optional[_Iterable[_Union[ConceptDictionaryTypeField, _Mapping]]]=..., description: _Optional[str]=...) -> None:
        ...

class ConceptDictionaryTypeField(_message.Message):
    __slots__ = ('key', 'value', 'description')
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    key: str
    value: ConceptType
    description: str

    def __init__(self, key: _Optional[str]=..., value: _Optional[_Union[ConceptType, _Mapping]]=..., description: _Optional[str]=...) -> None:
        ...

class ConceptTableType(_message.Message):
    __slots__ = ('columns', 'description')
    COLUMNS_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    columns: _containers.RepeatedCompositeFieldContainer[ConceptTableTypeColumn]
    description: str

    def __init__(self, columns: _Optional[_Iterable[_Union[ConceptTableTypeColumn, _Mapping]]]=..., description: _Optional[str]=...) -> None:
        ...

class ConceptTableTypeColumn(_message.Message):
    __slots__ = ('key', 'value', 'description')
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    key: str
    value: ConceptType
    description: str

    def __init__(self, key: _Optional[str]=..., value: _Optional[_Union[ConceptType, _Mapping]]=..., description: _Optional[str]=...) -> None:
        ...

class ConceptOpaqueType(_message.Message):
    __slots__ = ('is_a', 'description')
    IS_A_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    is_a: _containers.RepeatedCompositeFieldContainer[_noun_phrase_pb2.NounPhrase]
    description: str

    def __init__(self, is_a: _Optional[_Iterable[_Union[_noun_phrase_pb2.NounPhrase, _Mapping]]]=..., description: _Optional[str]=...) -> None:
        ...

class ConceptAnyType(_message.Message):
    __slots__ = ()

    def __init__(self) -> None:
        ...

class ConceptSelfType(_message.Message):
    __slots__ = ()

    def __init__(self) -> None:
        ...

class ConceptEnumTypeMember(_message.Message):
    __slots__ = ('name', 'noun_phrase', 'description')
    NAME_FIELD_NUMBER: _ClassVar[int]
    NOUN_PHRASE_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    name: str
    noun_phrase: _noun_phrase_pb2.NounPhrase
    description: str

    def __init__(self, name: _Optional[str]=..., noun_phrase: _Optional[_Union[_noun_phrase_pb2.NounPhrase, _Mapping]]=..., description: _Optional[str]=...) -> None:
        ...

class ConceptEnumType(_message.Message):
    __slots__ = ('is_a', 'members', 'description')
    IS_A_FIELD_NUMBER: _ClassVar[int]
    MEMBERS_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    is_a: _containers.RepeatedCompositeFieldContainer[_noun_phrase_pb2.NounPhrase]
    members: _containers.RepeatedCompositeFieldContainer[ConceptEnumTypeMember]
    description: str

    def __init__(self, is_a: _Optional[_Iterable[_Union[_noun_phrase_pb2.NounPhrase, _Mapping]]]=..., members: _Optional[_Iterable[_Union[ConceptEnumTypeMember, _Mapping]]]=..., description: _Optional[str]=...) -> None:
        ...

class ConceptType(_message.Message):
    __slots__ = ('scalar_type', 'optional_type', 'list_type', 'dictionary_type', 'table_type', 'opaque_type', 'any_type', 'union_type', 'self_type', 'sensitive_type', 'enum_type')
    SCALAR_TYPE_FIELD_NUMBER: _ClassVar[int]
    OPTIONAL_TYPE_FIELD_NUMBER: _ClassVar[int]
    LIST_TYPE_FIELD_NUMBER: _ClassVar[int]
    DICTIONARY_TYPE_FIELD_NUMBER: _ClassVar[int]
    TABLE_TYPE_FIELD_NUMBER: _ClassVar[int]
    OPAQUE_TYPE_FIELD_NUMBER: _ClassVar[int]
    ANY_TYPE_FIELD_NUMBER: _ClassVar[int]
    UNION_TYPE_FIELD_NUMBER: _ClassVar[int]
    SELF_TYPE_FIELD_NUMBER: _ClassVar[int]
    SENSITIVE_TYPE_FIELD_NUMBER: _ClassVar[int]
    ENUM_TYPE_FIELD_NUMBER: _ClassVar[int]
    scalar_type: ConceptScalarType
    optional_type: ConceptOptionalType
    list_type: ConceptListType
    dictionary_type: ConceptDictionaryType
    table_type: ConceptTableType
    opaque_type: ConceptOpaqueType
    any_type: ConceptAnyType
    union_type: ConceptUnionType
    self_type: ConceptSelfType
    sensitive_type: ConceptSensitiveType
    enum_type: ConceptEnumType

    def __init__(self, scalar_type: _Optional[_Union[ConceptScalarType, str]]=..., optional_type: _Optional[_Union[ConceptOptionalType, _Mapping]]=..., list_type: _Optional[_Union[ConceptListType, _Mapping]]=..., dictionary_type: _Optional[_Union[ConceptDictionaryType, _Mapping]]=..., table_type: _Optional[_Union[ConceptTableType, _Mapping]]=..., opaque_type: _Optional[_Union[ConceptOpaqueType, _Mapping]]=..., any_type: _Optional[_Union[ConceptAnyType, _Mapping]]=..., union_type: _Optional[_Union[ConceptUnionType, _Mapping]]=..., self_type: _Optional[_Union[ConceptSelfType, _Mapping]]=..., sensitive_type: _Optional[_Union[ConceptSensitiveType, _Mapping]]=..., enum_type: _Optional[_Union[ConceptEnumType, _Mapping]]=...) -> None:
        ...


from dataclasses import dataclass
from typing import Union


@dataclass
class AWSOffload:
    access_key: str
    secret_key: str
    session_token: str
    region: str
    bucket: str
    folder_name: str


Offload = Union[AWSOffload,]  # pyright: ignore [reportInvalidTypeArguments]



from bdk.v2.types import concept_type_pb2 as _concept_type_pb2
from bdk.v2.types import noun_phrase_pb2 as _noun_phrase_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union
DESCRIPTOR: _descriptor.FileDescriptor

class QuestionDescriptor(_message.Message):
    __slots__ = ('noun_phrases', 'type')
    NOUN_PHRASES_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    noun_phrases: _noun_phrase_pb2.NounPhrases
    type: _concept_type_pb2.ConceptType

    def __init__(self, noun_phrases: _Optional[_Union[_noun_phrase_pb2.NounPhrases, _Mapping]]=..., type: _Optional[_Union[_concept_type_pb2.ConceptType, _Mapping]]=...) -> None:
        ...


from abc import ABC, abstractmethod
from typing import Any, List, Optional, Tuple

from .book_descriptor import BookDescriptor
from .book_procedure_descriptor import BookProcedureDescriptor
from .concept_value import ConceptValue
from .discoverable import Discoverable
from .environment_information import EnvironmentInformation
from .expression import Expression
from .input_mappers.input_concepts import InputConcept
from .offload import Offload
from .promise import Promise
from .question_answer import AnsweredQuestion, Question
from .request_context import RequestContext

InvokeProcedureResponse = List[ConceptValue] | List[Question] | Promise


class TransportError(Exception):
    pass


class Transport(ABC):
    @abstractmethod
    def get_url(self) -> str:
        pass

    @abstractmethod
    def retrieve_books(
        self, context: Optional[RequestContext] = None
    ) -> List[BookDescriptor]:
        pass

    @abstractmethod
    def environment_information(
        self, context: Optional[RequestContext] = None
    ) -> EnvironmentInformation:
        pass

    @abstractmethod
    def retrieve_book(
        self, name: str, version: str, context: Optional[RequestContext] = None
    ) -> BookDescriptor:
        pass

    @abstractmethod
    def retrieve_procedures(
        self,
        name: str,
        version: str,
        include_connect: bool,
        context: Optional[RequestContext] = None,
    ) -> List[BookProcedureDescriptor]:
        pass

    @abstractmethod
    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def test_connection(
        self,
        name: str,
        version: str,
        authentication_id: Optional[str],
        authentication_credentials: Optional[List[Tuple[str, Any]]],
        context: Optional[RequestContext] = None,
    ) -> Any:
        pass

    @abstractmethod
    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def invoke_procedure(
        self,
        name: str,
        version: str,
        procedure_id: str,
        authentication_id: Optional[str],
        authentication_credentials: Optional[List[Tuple[str, Any]]],
        input_concepts: List[InputConcept],
        context: Optional[RequestContext] = None,
        filter_expression: Optional[Expression] = None,
        offload: Optional[Offload] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        answered_questions: Optional[List[AnsweredQuestion]] = None,
    ) -> InvokeProcedureResponse:
        pass

    @abstractmethod
    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def discover_procedures(
        self,
        name: str,
        version: str,
        authentication_id: Optional[str],
        authentication_credentials: Optional[List[Tuple[str, Any]]],
        what: List[str],
        context: Optional[RequestContext] = None,
    ) -> List[BookProcedureDescriptor]:
        pass

    @abstractmethod
    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def retrieve_discoverables(
        self,
        name: str,
        version: str,
        authentication_id: Optional[str],
        authentication_credentials: Optional[List[Tuple[str, Any]]],
        search: Optional[str],
        limit: Optional[int],
        offset: Optional[int],
        context: Optional[RequestContext] = None,
    ) -> List[Discoverable]:
        pass

    @abstractmethod
    def retrieve_tags(self, context: Optional[RequestContext] = None) -> List[str]:
        pass

    @abstractmethod
    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def resolve_promise(
        self,
        name: str,
        version: str,
        procedure_id: str,
        authentication_id: Optional[str],
        authentication_credentials: Optional[List[Tuple[str, Any]]],
        promise: Promise,
        configurations: List[InputConcept],
        offload: Optional[Offload] = None,
        answered_questions: Optional[List[AnsweredQuestion]] = None,
        context: Optional[RequestContext] = None,
    ) -> InvokeProcedureResponse:
        pass



from bdk.v2.types import noun_phrase_pb2 as _noun_phrase_pb2
from bdk.v2.types import value_pb2 as _value_pb2
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union
DESCRIPTOR: _descriptor.FileDescriptor

class BinaryOperator(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    BinaryOperatorAnd: _ClassVar[BinaryOperator]
    BinaryOperatorOr: _ClassVar[BinaryOperator]
    BinaryOperatorEquals: _ClassVar[BinaryOperator]
    BinaryOperatorNotEquals: _ClassVar[BinaryOperator]
    BinaryOperatorIn: _ClassVar[BinaryOperator]
    BinaryOperatorHas: _ClassVar[BinaryOperator]
    BinaryOperatorLessThan: _ClassVar[BinaryOperator]
    BinaryOperatorGreaterThan: _ClassVar[BinaryOperator]
    BinaryOperatorLessThanOrEqual: _ClassVar[BinaryOperator]
    BinaryOperatorGreaterThanOrEqual: _ClassVar[BinaryOperator]

class UnaryOperator(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    UnaryOperatorNot: _ClassVar[UnaryOperator]
BinaryOperatorAnd: BinaryOperator
BinaryOperatorOr: BinaryOperator
BinaryOperatorEquals: BinaryOperator
BinaryOperatorNotEquals: BinaryOperator
BinaryOperatorIn: BinaryOperator
BinaryOperatorHas: BinaryOperator
BinaryOperatorLessThan: BinaryOperator
BinaryOperatorGreaterThan: BinaryOperator
BinaryOperatorLessThanOrEqual: BinaryOperator
BinaryOperatorGreaterThanOrEqual: BinaryOperator
UnaryOperatorNot: UnaryOperator

class Expression(_message.Message):
    __slots__ = ('binary_expression', 'unary_expression', 'value_expression', 'noun_phrases_expression')
    BINARY_EXPRESSION_FIELD_NUMBER: _ClassVar[int]
    UNARY_EXPRESSION_FIELD_NUMBER: _ClassVar[int]
    VALUE_EXPRESSION_FIELD_NUMBER: _ClassVar[int]
    NOUN_PHRASES_EXPRESSION_FIELD_NUMBER: _ClassVar[int]
    binary_expression: BinaryExpression
    unary_expression: UnaryExpression
    value_expression: _value_pb2.Value
    noun_phrases_expression: _noun_phrase_pb2.NounPhrases

    def __init__(self, binary_expression: _Optional[_Union[BinaryExpression, _Mapping]]=..., unary_expression: _Optional[_Union[UnaryExpression, _Mapping]]=..., value_expression: _Optional[_Union[_value_pb2.Value, _Mapping]]=..., noun_phrases_expression: _Optional[_Union[_noun_phrase_pb2.NounPhrases, _Mapping]]=...) -> None:
        ...

class BinaryExpression(_message.Message):
    __slots__ = ('operator', 'left', 'right')
    OPERATOR_FIELD_NUMBER: _ClassVar[int]
    LEFT_FIELD_NUMBER: _ClassVar[int]
    RIGHT_FIELD_NUMBER: _ClassVar[int]
    operator: BinaryOperator
    left: Expression
    right: Expression

    def __init__(self, operator: _Optional[_Union[BinaryOperator, str]]=..., left: _Optional[_Union[Expression, _Mapping]]=..., right: _Optional[_Union[Expression, _Mapping]]=...) -> None:
        ...

class UnaryExpression(_message.Message):
    __slots__ = ('operator', 'inner')
    OPERATOR_FIELD_NUMBER: _ClassVar[int]
    INNER_FIELD_NUMBER: _ClassVar[int]
    operator: UnaryOperator
    inner: Expression

    def __init__(self, operator: _Optional[_Union[UnaryOperator, str]]=..., inner: _Optional[_Union[Expression, _Mapping]]=...) -> None:
        ...


# pylint: disable=no-name-in-module
from datetime import timedelta, timezone

import google.protobuf.timestamp_pb2

from ..proto.bdk.v1.types.date_pb2 import Date as ProtoDate
from ..proto.bdk.v1.types.time_pb2 import Time as ProtoTime
from ..proto.bdk.v1.types.value_pb2 import DictionaryValue as ProtoDict
from ..proto.bdk.v1.types.value_pb2 import FileValue as ProtoFileValue
from ..proto.bdk.v1.types.value_pb2 import ListValue as ProtoList
from ..proto.bdk.v1.types.value_pb2 import OpaqueValue as ProtoOpaqueValue
from ..proto.bdk.v1.types.value_pb2 import SensitiveValue as ProtoSensitive
from ..proto.bdk.v1.types.value_pb2 import TableValue as ProtoTableValue
from ..proto.bdk.v1.types.value_pb2 import Value as ProtoValue
from ..value import (BooleanValue, Column, DatetimeValue, DateValue,
                     DictionaryValue, File, NullValue, NumberValue,
                     OpaqueValue, RemoteFile, RemoteTable, SensitiveValue,
                     Table, TextValue, TimeValue, Value)
from . import noun_phrase as noun_phrase_mapper


def _nanos_to_micros(nanos: int) -> int:
    return nanos // 1000


def _convert_timestamp(
    proto_timestamp: google.protobuf.timestamp_pb2.Timestamp,  # pylint: disable=no-member
) -> DatetimeValue:
    datetime = DatetimeValue.fromtimestamp(proto_timestamp.seconds, tz=timezone.utc)

    return datetime + timedelta(microseconds=_nanos_to_micros(proto_timestamp.nanos))


def _convert_date(proto_date: ProtoDate):
    return DateValue(year=proto_date.year, month=proto_date.month, day=proto_date.day)


def _convert_time(proto_time: ProtoTime):
    return TimeValue(
        hour=proto_time.hours,
        minute=proto_time.minutes,
        second=proto_time.seconds,
        microsecond=_nanos_to_micros(proto_time.nanos),
    )


def _convert_list(proto_list: ProtoList):
    return list(map(map_value, proto_list.values))


def _convert_sensitive(proto_sensitive: ProtoSensitive):
    return SensitiveValue(map_value(proto_sensitive.value))


def _convert_dictionary(proto_dict: ProtoDict):
    return DictionaryValue(
        value={f.key: map_value(f.value) for f in proto_dict.fields},
        is_a=[noun_phrase_mapper.map_noun_phrase(is_a) for is_a in proto_dict.is_a],
    )


def _convert_file(proto_file: ProtoFileValue):
    file_type = proto_file.WhichOneof("inline_remote")

    if file_type == "remote":
        return RemoteFile(url=proto_file.remote)

    if file_type == "inline":
        return File(
            filename=proto_file.inline.file_name, content=proto_file.inline.content
        )

    raise NotImplementedError(f"Could not map file value `{proto_file}`")


def _convert_opaque(proto_opaque: ProtoOpaqueValue):
    return OpaqueValue(
        is_a=[noun_phrase_mapper.map_noun_phrase(is_a) for is_a in proto_opaque.is_a],
        value=proto_opaque.content,
    )


def _convert_table(proto_table: ProtoTableValue):
    proto_table_type = proto_table.WhichOneof("inline_remote")

    if proto_table_type == "inline":
        return Table(
            columns=[
                Column(name=c.name, values=[map_value(val) for val in c.values])
                for c in proto_table.inline.columns
            ],
        )

    if proto_table_type == "remote":
        return RemoteTable(url=proto_table.remote)

    raise NotImplementedError(f"Could not map table value `{proto_table}`")


def map_value(value: ProtoValue) -> Value:
    value_type = value.WhichOneof("value_discriminator")

    if not value_type:
        return None

    value_type_mappings = {
        "conceptual_value": noun_phrase_mapper.map_noun_phrase,
        "number_value": NumberValue,
        "null_value": lambda v: NullValue(),
        "text_value": TextValue,
        "boolean_value": BooleanValue,
        "opaque_value": _convert_opaque,
        "datetime_value": _convert_timestamp,
        "date_value": _convert_date,
        "time_value": _convert_time,
        "list_value": _convert_list,
        "dictionary_value": _convert_dictionary,
        "file_value": _convert_file,
        "table_value": _convert_table,
        "sensitive_value": _convert_sensitive,
    }

    if value_type not in value_type_mappings:
        raise NotImplementedError(f"Could not map value `{value}`")

    return value_type_mappings[value_type](getattr(value, value_type))


def dict_cast(value: dict):
    return DictionaryValue(
        value=value,
        is_a=[],
    )



from dataclasses import dataclass
from enum import Enum
from typing import List, Union

from .base import BdkDescriptorBase
from .noun_phrase import NounPhrase


@dataclass
class ConceptTableTypeColumn(BdkDescriptorBase):
    key: str
    value: "ConceptType"


class ConceptScalarType(Enum):
    CONCEPT_SCALAR_TYPE_CONCEPTUAL = 0
    CONCEPT_SCALAR_TYPE_TEXT = 1
    CONCEPT_SCALAR_TYPE_NUMBER = 2
    CONCEPT_SCALAR_TYPE_BOOLEAN = 3
    CONCEPT_SCALAR_TYPE_DATETIME = 4
    CONCEPT_SCALAR_TYPE_DATE = 5
    CONCEPT_SCALAR_TYPE_TIME = 6
    CONCEPT_SCALAR_TYPE_FILE = 7
    CONCEPT_SCALAR_TYPE_UUID = 8

    def human_readable(self) -> str:
        return self.name.replace("CONCEPT_SCALAR_TYPE_", "")


@dataclass
class ConceptOptionalType(BdkDescriptorBase):
    inner: "ConceptType"


@dataclass
class ConceptSensitiveType(BdkDescriptorBase):
    inner: "ConceptType"


@dataclass
class ConceptListType(BdkDescriptorBase):
    inner: "ConceptType"


@dataclass
class ConceptDictionaryTypeField(BdkDescriptorBase):
    key: str
    value: "ConceptType"
    description: str


@dataclass
class ConceptDictionaryType(BdkDescriptorBase):
    is_a: List[NounPhrase]
    fields: List[ConceptDictionaryTypeField]
    description: str

    def human_readable(self) -> str:
        """
        Returns a string representation of the DictionaryValue.
        """
        str_is_a = [str(is_a) for is_a in self.is_a]
        return f"Object({', '.join(str_is_a)})"


@dataclass
class ConceptTableType(BdkDescriptorBase):
    columns: List[ConceptTableTypeColumn]
    description: str

    def human_readable(self) -> str:
        """
        Returns a string representation of a Table.
        """
        str_columns = [column.key for column in self.columns]
        return f"Table with columns: {', '.join(str_columns)}"


@dataclass
class ConceptOpaqueType(BdkDescriptorBase):
    is_a: List[NounPhrase]
    description: str

    def human_readable(self) -> str:
        """
        Returns a string representation of the OpaqueValue.
        """
        str_is_a = [str(is_a) for is_a in self.is_a]
        return f"Object({', '.join(str_is_a)})"


@dataclass
class ConceptAnyType(BdkDescriptorBase):
    pass


@dataclass
class ConceptSelfType(BdkDescriptorBase):
    pass


@dataclass
class ConceptUnionType(BdkDescriptorBase):
    inners: List["ConceptType"]


@dataclass
class ConceptEnumTypeMember(BdkDescriptorBase):
    name: str
    noun_phrase: NounPhrase
    description: str


@dataclass
class ConceptEnumType(BdkDescriptorBase):
    is_a: List[NounPhrase]
    members: List[ConceptEnumTypeMember]
    description: str

    def human_readable(self) -> str:
        """
        Returns a string representation of the EnumValue.
        """
        str_is_a = [str(is_a) for is_a in self.is_a]
        return f"Enum({', '.join(str_is_a)})"


ConceptType = Union[
    ConceptScalarType,
    ConceptOptionalType,
    ConceptListType,
    ConceptDictionaryType,
    ConceptTableType,
    ConceptOpaqueType,
    ConceptAnyType,
    ConceptUnionType,
    ConceptSelfType,
    ConceptSensitiveType,
    ConceptEnumType,
]



from bdk.v1.types import book_connection_descriptor_pb2 as _book_connection_descriptor_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union
DESCRIPTOR: _descriptor.FileDescriptor

class BookConnectionResponse(_message.Message):
    __slots__ = ('book_connection_descriptor', 'oauth_redirect')
    BOOK_CONNECTION_DESCRIPTOR_FIELD_NUMBER: _ClassVar[int]
    OAUTH_REDIRECT_FIELD_NUMBER: _ClassVar[int]
    book_connection_descriptor: _book_connection_descriptor_pb2.BookConnectionDescriptor
    oauth_redirect: BookConnectionOAuthRedirectResponse

    def __init__(self, book_connection_descriptor: _Optional[_Union[_book_connection_descriptor_pb2.BookConnectionDescriptor, _Mapping]]=..., oauth_redirect: _Optional[_Union[BookConnectionOAuthRedirectResponse, _Mapping]]=...) -> None:
        ...

class BookConnectionOAuthRedirectResponse(_message.Message):
    __slots__ = ('url',)
    URL_FIELD_NUMBER: _ClassVar[int]
    url: str

    def __init__(self, url: _Optional[str]=...) -> None:
        ...


"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_sym_db = _symbol_database.Default()
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$bdk/v1/requests/retrieve_books.proto\x12\x06bdk.v1"y\n\x14RetrieveBooksRequest\x12 \n\tpage_size\x18\x01 \x01(\rH\x00R\x08pageSize\x88\x01\x01\x12"\n\npage_token\x18\x02 \x01(\tH\x01R\tpageToken\x88\x01\x01B\x0c\n\n_page_sizeB\r\n\x0b_page_tokenB\x1fZ\x1dgithub.com/kognitos/bdk-protob\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'bdk.v1.requests.retrieve_books_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals['DESCRIPTOR']._options = None
    _globals['DESCRIPTOR']._serialized_options = b'Z\x1dgithub.com/kognitos/bdk-proto'
    _globals['_RETRIEVEBOOKSREQUEST']._serialized_start = 48
    _globals['_RETRIEVEBOOKSREQUEST']._serialized_end = 169


from kognitos.bdk.runtime.client.concept_descriptor import ConceptDescriptor
# pylint: disable=no-name-in-module
from kognitos.bdk.runtime.client.proto.bdk.v1.types.concept_descriptor_pb2 import \
    ConceptDescriptor as ProtoConceptDescriptor

from . import concept_type as ct
from . import noun_phrase as np
from . import value as v


def map_concept_descriptor(data: ProtoConceptDescriptor) -> ConceptDescriptor:
    return ConceptDescriptor(
        description=data.description,
        type=ct.map_concept_type(data.type),
        noun_phrases=np.map_noun_phrases(data.noun_phrases),
        default_value=v.map_value(data.default_value),
    )



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_sym_db = _symbol_database.Default()
from ....bdk.v1.types import noun_phrase_pb2 as bdk_dot_v1_dot_types_dot_noun__phrase__pb2
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1fbdk/v1/types/concept_type.proto\x12\x06bdk.v1\x1a\x1ebdk/v1/types/noun_phrase.proto"@\n\x13ConceptOptionalType\x12)\n\x05inner\x18\x01 \x01(\x0b2\x13.bdk.v1.ConceptTypeR\x05inner"A\n\x14ConceptSensitiveType\x12)\n\x05inner\x18\x01 \x01(\x0b2\x13.bdk.v1.ConceptTypeR\x05inner"<\n\x0fConceptListType\x12)\n\x05inner\x18\x01 \x01(\x0b2\x13.bdk.v1.ConceptTypeR\x05inner"?\n\x10ConceptUnionType\x12+\n\x06inners\x18\x01 \x03(\x0b2\x13.bdk.v1.ConceptTypeR\x06inners"\xb1\x01\n\x15ConceptDictionaryType\x12%\n\x04is_a\x18\x01 \x03(\x0b2\x12.bdk.v1.NounPhraseR\x03isA\x12:\n\x06fields\x18\x02 \x03(\x0b2".bdk.v1.ConceptDictionaryTypeFieldR\x06fields\x12%\n\x0bdescription\x18\x03 \x01(\tH\x00R\x0bdescription\x88\x01\x01B\x0e\n\x0c_description"\x90\x01\n\x1aConceptDictionaryTypeField\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12)\n\x05value\x18\x02 \x01(\x0b2\x13.bdk.v1.ConceptTypeR\x05value\x12%\n\x0bdescription\x18\x03 \x01(\tH\x00R\x0bdescription\x88\x01\x01B\x0e\n\x0c_description"\x83\x01\n\x10ConceptTableType\x128\n\x07columns\x18\x01 \x03(\x0b2\x1e.bdk.v1.ConceptTableTypeColumnR\x07columns\x12%\n\x0bdescription\x18\x02 \x01(\tH\x00R\x0bdescription\x88\x01\x01B\x0e\n\x0c_description"\x8c\x01\n\x16ConceptTableTypeColumn\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12)\n\x05value\x18\x02 \x01(\x0b2\x13.bdk.v1.ConceptTypeR\x05value\x12%\n\x0bdescription\x18\x03 \x01(\tH\x00R\x0bdescription\x88\x01\x01B\x0e\n\x0c_description"q\n\x11ConceptOpaqueType\x12%\n\x04is_a\x18\x01 \x03(\x0b2\x12.bdk.v1.NounPhraseR\x03isA\x12%\n\x0bdescription\x18\x02 \x01(\tH\x00R\x0bdescription\x88\x01\x01B\x0e\n\x0c_description"\x10\n\x0eConceptAnyType"\x11\n\x0fConceptSelfType"\x97\x01\n\x15ConceptEnumTypeMember\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x123\n\x0bnoun_phrase\x18\x02 \x01(\x0b2\x12.bdk.v1.NounPhraseR\nnounPhrase\x12%\n\x0bdescription\x18\x03 \x01(\tH\x00R\x0bdescription\x88\x01\x01B\x0e\n\x0c_description"\xa8\x01\n\x0fConceptEnumType\x12%\n\x04is_a\x18\x01 \x03(\x0b2\x12.bdk.v1.NounPhraseR\x03isA\x127\n\x07members\x18\x02 \x03(\x0b2\x1d.bdk.v1.ConceptEnumTypeMemberR\x07members\x12%\n\x0bdescription\x18\x03 \x01(\tH\x00R\x0bdescription\x88\x01\x01B\x0e\n\x0c_description"\xcf\x05\n\x0bConceptType\x12<\n\x0bscalar_type\x18\x01 \x01(\x0e2\x19.bdk.v1.ConceptScalarTypeH\x00R\nscalarType\x12B\n\roptional_type\x18\x02 \x01(\x0b2\x1b.bdk.v1.ConceptOptionalTypeH\x00R\x0coptionalType\x126\n\tlist_type\x18\x03 \x01(\x0b2\x17.bdk.v1.ConceptListTypeH\x00R\x08listType\x12H\n\x0fdictionary_type\x18\x04 \x01(\x0b2\x1d.bdk.v1.ConceptDictionaryTypeH\x00R\x0edictionaryType\x129\n\ntable_type\x18\x05 \x01(\x0b2\x18.bdk.v1.ConceptTableTypeH\x00R\ttableType\x12<\n\x0bopaque_type\x18\x06 \x01(\x0b2\x19.bdk.v1.ConceptOpaqueTypeH\x00R\nopaqueType\x123\n\x08any_type\x18\x07 \x01(\x0b2\x16.bdk.v1.ConceptAnyTypeH\x00R\x07anyType\x129\n\nunion_type\x18\x08 \x01(\x0b2\x18.bdk.v1.ConceptUnionTypeH\x00R\tunionType\x126\n\tself_type\x18\t \x01(\x0b2\x17.bdk.v1.ConceptSelfTypeH\x00R\x08selfType\x12E\n\x0esensitive_type\x18\n \x01(\x0b2\x1c.bdk.v1.ConceptSensitiveTypeH\x00R\rsensitiveType\x126\n\tenum_type\x18\x0b \x01(\x0b2\x17.bdk.v1.ConceptEnumTypeH\x00R\x08enumTypeB\x1c\n\x1aconcept_type_discriminator*\x95\x02\n\x11ConceptScalarType\x12\x1f\n\x1bConceptScalarTypeConceptual\x10\x00\x12\x19\n\x15ConceptScalarTypeText\x10\x01\x12\x1b\n\x17ConceptScalarTypeNumber\x10\x02\x12\x1c\n\x18ConceptScalarTypeBoolean\x10\x03\x12\x1d\n\x19ConceptScalarTypeDatetime\x10\x04\x12\x19\n\x15ConceptScalarTypeDate\x10\x05\x12\x19\n\x15ConceptScalarTypeTime\x10\x06\x12\x19\n\x15ConceptScalarTypeFile\x10\x07\x12\x19\n\x15ConceptScalarTypeUUID\x10\x08B\x1fZ\x1dgithub.com/kognitos/bdk-protob\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'bdk.v1.types.concept_type_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals['DESCRIPTOR']._options = None
    _globals['DESCRIPTOR']._serialized_options = b'Z\x1dgithub.com/kognitos/bdk-proto'
    _globals['_CONCEPTSCALARTYPE']._serialized_start = 2139
    _globals['_CONCEPTSCALARTYPE']._serialized_end = 2416
    _globals['_CONCEPTOPTIONALTYPE']._serialized_start = 75
    _globals['_CONCEPTOPTIONALTYPE']._serialized_end = 139
    _globals['_CONCEPTSENSITIVETYPE']._serialized_start = 141
    _globals['_CONCEPTSENSITIVETYPE']._serialized_end = 206
    _globals['_CONCEPTLISTTYPE']._serialized_start = 208
    _globals['_CONCEPTLISTTYPE']._serialized_end = 268
    _globals['_CONCEPTUNIONTYPE']._serialized_start = 270
    _globals['_CONCEPTUNIONTYPE']._serialized_end = 333
    _globals['_CONCEPTDICTIONARYTYPE']._serialized_start = 336
    _globals['_CONCEPTDICTIONARYTYPE']._serialized_end = 513
    _globals['_CONCEPTDICTIONARYTYPEFIELD']._serialized_start = 516
    _globals['_CONCEPTDICTIONARYTYPEFIELD']._serialized_end = 660
    _globals['_CONCEPTTABLETYPE']._serialized_start = 663
    _globals['_CONCEPTTABLETYPE']._serialized_end = 794
    _globals['_CONCEPTTABLETYPECOLUMN']._serialized_start = 797
    _globals['_CONCEPTTABLETYPECOLUMN']._serialized_end = 937
    _globals['_CONCEPTOPAQUETYPE']._serialized_start = 939
    _globals['_CONCEPTOPAQUETYPE']._serialized_end = 1052
    _globals['_CONCEPTANYTYPE']._serialized_start = 1054
    _globals['_CONCEPTANYTYPE']._serialized_end = 1070
    _globals['_CONCEPTSELFTYPE']._serialized_start = 1072
    _globals['_CONCEPTSELFTYPE']._serialized_end = 1089
    _globals['_CONCEPTENUMTYPEMEMBER']._serialized_start = 1092
    _globals['_CONCEPTENUMTYPEMEMBER']._serialized_end = 1243
    _globals['_CONCEPTENUMTYPE']._serialized_start = 1246
    _globals['_CONCEPTENUMTYPE']._serialized_end = 1414
    _globals['_CONCEPTTYPE']._serialized_start = 1417
    _globals['_CONCEPTTYPE']._serialized_end = 2136


"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_sym_db = _symbol_database.Default()
from ....bdk.v1.requests import add_book_connection_pb2 as bdk_dot_v1_dot_requests_dot_add__book__connection__pb2
from ....bdk.v1.requests import book_connection_oauth_callback_pb2 as bdk_dot_v1_dot_requests_dot_book__connection__oauth__callback__pb2
from ....bdk.v1.requests import discover_procedures_pb2 as bdk_dot_v1_dot_requests_dot_discover__procedures__pb2
from ....bdk.v1.requests import environment_information_pb2 as bdk_dot_v1_dot_requests_dot_environment__information__pb2
from ....bdk.v1.requests import invoke_procedure_pb2 as bdk_dot_v1_dot_requests_dot_invoke__procedure__pb2
from ....bdk.v1.requests import remove_book_connection_pb2 as bdk_dot_v1_dot_requests_dot_remove__book__connection__pb2
from ....bdk.v1.requests import retrieve_book_pb2 as bdk_dot_v1_dot_requests_dot_retrieve__book__pb2
from ....bdk.v1.requests import retrieve_book_connections_pb2 as bdk_dot_v1_dot_requests_dot_retrieve__book__connections__pb2
from ....bdk.v1.requests import retrieve_book_procedures_pb2 as bdk_dot_v1_dot_requests_dot_retrieve__book__procedures__pb2
from ....bdk.v1.requests import retrieve_books_pb2 as bdk_dot_v1_dot_requests_dot_retrieve__books__pb2
from ....bdk.v1.requests import retrieve_books_connections_pb2 as bdk_dot_v1_dot_requests_dot_retrieve__books__connections__pb2
from ....bdk.v1.requests import retrieve_tags_pb2 as bdk_dot_v1_dot_requests_dot_retrieve__tags__pb2
from ....bdk.v1.requests import test_connection_pb2 as bdk_dot_v1_dot_requests_dot_test__connection__pb2
from ....bdk.v1.requests import update_book_connection_pb2 as bdk_dot_v1_dot_requests_dot_update__book__connection__pb2
from ....bdk.v1.requests import upgrade_book_connection_pb2 as bdk_dot_v1_dot_requests_dot_upgrade__book__connection__pb2
from ....bdk.v1.responses import book_connection_pb2 as bdk_dot_v1_dot_responses_dot_book__connection__pb2
from ....bdk.v1.responses import discover_procedures_pb2 as bdk_dot_v1_dot_responses_dot_discover__procedures__pb2
from ....bdk.v1.responses import environment_information_pb2 as bdk_dot_v1_dot_responses_dot_environment__information__pb2
from ....bdk.v1.responses import invoke_procedure_pb2 as bdk_dot_v1_dot_responses_dot_invoke__procedure__pb2
from ....bdk.v1.responses import remove_book_connection_pb2 as bdk_dot_v1_dot_responses_dot_remove__book__connection__pb2
from ....bdk.v1.responses import retrieve_book_pb2 as bdk_dot_v1_dot_responses_dot_retrieve__book__pb2
from ....bdk.v1.responses import retrieve_book_connections_pb2 as bdk_dot_v1_dot_responses_dot_retrieve__book__connections__pb2
from ....bdk.v1.responses import retrieve_book_procedures_pb2 as bdk_dot_v1_dot_responses_dot_retrieve__book__procedures__pb2
from ....bdk.v1.responses import retrieve_books_pb2 as bdk_dot_v1_dot_responses_dot_retrieve__books__pb2
from ....bdk.v1.responses import retrieve_books_connections_pb2 as bdk_dot_v1_dot_responses_dot_retrieve__books__connections__pb2
from ....bdk.v1.responses import retrieve_tags_pb2 as bdk_dot_v1_dot_responses_dot_retrieve__tags__pb2
from ....bdk.v1.responses import test_connection_pb2 as bdk_dot_v1_dot_responses_dot_test__connection__pb2
from ....bdk.v1.responses import upgrade_book_connection_pb2 as bdk_dot_v1_dot_responses_dot_upgrade__book__connection__pb2
from ....google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dbdk/v1/services/library.proto\x12\x06bdk.v1\x1a)bdk/v1/requests/add_book_connection.proto\x1a4bdk/v1/requests/book_connection_oauth_callback.proto\x1a)bdk/v1/requests/discover_procedures.proto\x1a-bdk/v1/requests/environment_information.proto\x1a&bdk/v1/requests/invoke_procedure.proto\x1a,bdk/v1/requests/remove_book_connection.proto\x1a#bdk/v1/requests/retrieve_book.proto\x1a/bdk/v1/requests/retrieve_book_connections.proto\x1a.bdk/v1/requests/retrieve_book_procedures.proto\x1a$bdk/v1/requests/retrieve_books.proto\x1a0bdk/v1/requests/retrieve_books_connections.proto\x1a#bdk/v1/requests/retrieve_tags.proto\x1a%bdk/v1/requests/test_connection.proto\x1a,bdk/v1/requests/update_book_connection.proto\x1a-bdk/v1/requests/upgrade_book_connection.proto\x1a&bdk/v1/responses/book_connection.proto\x1a*bdk/v1/responses/discover_procedures.proto\x1a.bdk/v1/responses/environment_information.proto\x1a\'bdk/v1/responses/invoke_procedure.proto\x1a-bdk/v1/responses/remove_book_connection.proto\x1a$bdk/v1/responses/retrieve_book.proto\x1a0bdk/v1/responses/retrieve_book_connections.proto\x1a/bdk/v1/responses/retrieve_book_procedures.proto\x1a%bdk/v1/responses/retrieve_books.proto\x1a1bdk/v1/responses/retrieve_books_connections.proto\x1a$bdk/v1/responses/retrieve_tags.proto\x1a&bdk/v1/responses/test_connection.proto\x1a.bdk/v1/responses/upgrade_book_connection.proto\x1a\x1cgoogle/api/annotations.proto2\x83\x0e\n\x0eLibraryService\x12c\n\rRetrieveBooks\x12\x1c.bdk.v1.RetrieveBooksRequest\x1a\x1d.bdk.v1.RetrieveBooksResponse"\x15\x82\xd3\xe4\x93\x02\x0f\x12\r/api/v1/books\x12q\n\x0cRetrieveBook\x12\x1b.bdk.v1.RetrieveBookRequest\x1a\x1c.bdk.v1.RetrieveBookResponse"&\x82\xd3\xe4\x93\x02 \x12\x1e/api/v1/books/{name}/{version}\x12\x9c\x01\n\x16RetrieveBookProcedures\x12%.bdk.v1.RetrieveBookProceduresRequest\x1a(.bdk.v1.RetrieveBookProceduresResponseV2"1\x82\xd3\xe4\x93\x02+\x12)/api/v1/books/{name}/{version}/procedures\x12\x97\x01\n\x0eTestConnection\x12\x1d.bdk.v1.TestConnectionRequest\x1a\x1e.bdk.v1.TestConnectionResponse"F\x82\xd3\xe4\x93\x02@"./api/v1/books/{name}/{version}/test-connection:\x0eauthentication\x12_\n\x0cRetrieveTags\x12\x1b.bdk.v1.RetrieveTagsRequest\x1a\x1c.bdk.v1.RetrieveTagsResponse"\x14\x82\xd3\xe4\x93\x02\x0e\x12\x0c/api/v1/tags\x12\x96\x01\n\x11AddBookConnection\x12 .bdk.v1.AddBookConnectionRequest\x1a\x1e.bdk.v1.BookConnectionResponse"?\x82\xd3\xe4\x93\x029"4/api/v1/books/{book_name}/{book_version}/connections:\x01*\x12\xac\x01\n\x14UpdateBookConnection\x12#.bdk.v1.UpdateBookConnectionRequest\x1a\x1e.bdk.v1.BookConnectionResponse"O\x82\xd3\xe4\x93\x02I\x1aD/api/v1/books/{book_name}/{book_version}/connections/{connection_id}:\x01*\x12\x8c\x01\n\x1bBookConnectionOAuthCallback\x12*.bdk.v1.BookConnectionOAuthCallbackRequest\x1a\x1e.bdk.v1.BookConnectionResponse"!\x82\xd3\xe4\x93\x02\x1b"\x16/api/v1/oauth/callback:\x01*\x12\xaf\x01\n\x14RemoveBookConnection\x12#.bdk.v1.RemoveBookConnectionRequest\x1a$.bdk.v1.RemoveBookConnectionResponse"L\x82\xd3\xe4\x93\x02F*D/api/v1/books/{book_name}/{book_version}/connections/{connection_id}\x12\xa8\x01\n\x17RetrieveBookConnections\x12&.bdk.v1.RetrieveBookConnectionsRequest\x1a\'.bdk.v1.RetrieveBookConnectionsResponse"<\x82\xd3\xe4\x93\x026\x124/api/v1/books/{book_name}/{book_version}/connections\x12\x8a\x01\n\x18RetrieveBooksConnections\x12\'.bdk.v1.RetrieveBooksConnectionsRequest\x1a(.bdk.v1.RetrieveBooksConnectionsResponse"\x1b\x82\xd3\xe4\x93\x02\x15\x12\x13/api/v1/connections\x12\xbd\x01\n\x15UpgradeBookConnection\x12$.bdk.v1.UpgradeBookConnectionRequest\x1a%.bdk.v1.UpgradeBookConnectionResponse"W\x82\xd3\xe4\x93\x02Q\x1aL/api/v1/books/{book_name}/{book_version}/connections/{connection_id}/upgrade:\x01*B\x1fZ\x1dgithub.com/kognitos/bdk-protob\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'bdk.v1.services.library_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals['DESCRIPTOR']._options = None
    _globals['DESCRIPTOR']._serialized_options = b'Z\x1dgithub.com/kognitos/bdk-proto'
    _globals['_LIBRARYSERVICE'].methods_by_name['RetrieveBooks']._options = None
    _globals['_LIBRARYSERVICE'].methods_by_name['RetrieveBooks']._serialized_options = b'\x82\xd3\xe4\x93\x02\x0f\x12\r/api/v1/books'
    _globals['_LIBRARYSERVICE'].methods_by_name['RetrieveBook']._options = None
    _globals['_LIBRARYSERVICE'].methods_by_name['RetrieveBook']._serialized_options = b'\x82\xd3\xe4\x93\x02 \x12\x1e/api/v1/books/{name}/{version}'
    _globals['_LIBRARYSERVICE'].methods_by_name['RetrieveBookProcedures']._options = None
    _globals['_LIBRARYSERVICE'].methods_by_name['RetrieveBookProcedures']._serialized_options = b'\x82\xd3\xe4\x93\x02+\x12)/api/v1/books/{name}/{version}/procedures'
    _globals['_LIBRARYSERVICE'].methods_by_name['TestConnection']._options = None
    _globals['_LIBRARYSERVICE'].methods_by_name['TestConnection']._serialized_options = b'\x82\xd3\xe4\x93\x02@"./api/v1/books/{name}/{version}/test-connection:\x0eauthentication'
    _globals['_LIBRARYSERVICE'].methods_by_name['RetrieveTags']._options = None
    _globals['_LIBRARYSERVICE'].methods_by_name['RetrieveTags']._serialized_options = b'\x82\xd3\xe4\x93\x02\x0e\x12\x0c/api/v1/tags'
    _globals['_LIBRARYSERVICE'].methods_by_name['AddBookConnection']._options = None
    _globals['_LIBRARYSERVICE'].methods_by_name['AddBookConnection']._serialized_options = b'\x82\xd3\xe4\x93\x029"4/api/v1/books/{book_name}/{book_version}/connections:\x01*'
    _globals['_LIBRARYSERVICE'].methods_by_name['UpdateBookConnection']._options = None
    _globals['_LIBRARYSERVICE'].methods_by_name['UpdateBookConnection']._serialized_options = b'\x82\xd3\xe4\x93\x02I\x1aD/api/v1/books/{book_name}/{book_version}/connections/{connection_id}:\x01*'
    _globals['_LIBRARYSERVICE'].methods_by_name['BookConnectionOAuthCallback']._options = None
    _globals['_LIBRARYSERVICE'].methods_by_name['BookConnectionOAuthCallback']._serialized_options = b'\x82\xd3\xe4\x93\x02\x1b"\x16/api/v1/oauth/callback:\x01*'
    _globals['_LIBRARYSERVICE'].methods_by_name['RemoveBookConnection']._options = None
    _globals['_LIBRARYSERVICE'].methods_by_name['RemoveBookConnection']._serialized_options = b'\x82\xd3\xe4\x93\x02F*D/api/v1/books/{book_name}/{book_version}/connections/{connection_id}'
    _globals['_LIBRARYSERVICE'].methods_by_name['RetrieveBookConnections']._options = None
    _globals['_LIBRARYSERVICE'].methods_by_name['RetrieveBookConnections']._serialized_options = b'\x82\xd3\xe4\x93\x026\x124/api/v1/books/{book_name}/{book_version}/connections'
    _globals['_LIBRARYSERVICE'].methods_by_name['RetrieveBooksConnections']._options = None
    _globals['_LIBRARYSERVICE'].methods_by_name['RetrieveBooksConnections']._serialized_options = b'\x82\xd3\xe4\x93\x02\x15\x12\x13/api/v1/connections'
    _globals['_LIBRARYSERVICE'].methods_by_name['UpgradeBookConnection']._options = None
    _globals['_LIBRARYSERVICE'].methods_by_name['UpgradeBookConnection']._serialized_options = b'\x82\xd3\xe4\x93\x02Q\x1aL/api/v1/books/{book_name}/{book_version}/connections/{connection_id}/upgrade:\x01*'
    _globals['_LIBRARYSERVICE']._serialized_start = 1309
    _globals['_LIBRARYSERVICE']._serialized_end = 3104


from .answered_question import map_answered_questions
from .authentication import map_authentication
from .expression import map_expression
from .input_concepts import map_input_concepts
from .offload import map_offload
from .promise import map_promise



import base64
from typing import Optional

import requests

from .base_transport import BaseTransport
from .proto.bdk.v1.requests.request_pb2 import \
    Request  # pylint: disable=no-name-in-module
from .proto.bdk.v1.responses.response_pb2 import \
    Response  # pylint: disable=no-name-in-module
from .transport import TransportError


class LambdaRIETransport(BaseTransport):
    def __init__(self, host: str = "localhost", port: Optional[int] = None):

        if host == "localhost":
            if port is None:
                port = 9000

        if not port:
            self.base_url = f"http://{host}/2015-03-31/functions/function/invocations"
        else:
            self.base_url = (
                f"http://{host}:{port}/2015-03-31/functions/function/invocations"
            )

    def get_url(self) -> str:
        return self.base_url

    def send(self, request: Request) -> Response:
        data_str = request.SerializeToString()

        data = base64.b64encode(data_str).decode("utf-8")

        response = requests.post(self.base_url, json=data, timeout=60)

        response.raise_for_status()

        try:
            response_payload = base64.b64decode(response.json())

            response = Response()
            response.ParseFromString(response_payload)

            return response
        except Exception as e:
            if "response_payload" in locals():
                response = f"response_payload: {locals().get('response_payload')}"
            raise TransportError(
                f"Could not decode the response <<<{response}>>> got by calling(LambdaRIETransport) <<<{self.base_url}>>> with the request <<<{request.WhichOneof('request_discriminator')}>>>"
            ) from e



import builtins
from dataclasses import dataclass
from functools import wraps
from logging import Logger
from typing import Any, Set

from kognitos.bdk.runtime.client.proto.bdk.v1.types.value_pb2 import \
    SensitiveValue as ProtoSensitiveValue  # pylint: disable=no-name-in-module

sensitive_data: Set[str] = set()


@dataclass
class Sensitive:
    def __init__(self, value: Any):
        sensitive_data.add(str(value))
        self.value = value

    def __str__(self):
        return "Sensitive(****)"

    def __repr__(self):
        return "Sensitive(****)"

    def __eq__(self, other):
        if isinstance(other, Sensitive):
            return self.value == other.value
        return False

    def __hash__(self):
        return hash(("Sensitive", self.value))


unwrapped_log_method = Logger._log
unwrapped_print_function = builtins.print


def scrub_filter(data: str):
    for sensitive in sorted(list(sensitive_data), reverse=True):
        data = data.replace(sensitive, "Sensitive(****)")
    return data


# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
@wraps(unwrapped_log_method)
def scrub_sensitive_data_from_logger(
    self, level, msg, args, exc_info=None, extra=None, stack_info=False, stacklevel=2
):
    scrubbed_msg = scrub_filter(str(msg))
    return unwrapped_log_method(
        self,
        level,
        scrubbed_msg,
        args,
        exc_info=exc_info,
        extra=extra,
        stack_info=stack_info,
        stacklevel=stacklevel,
    )


@wraps(builtins.print)
def scrub_sensitive_data_from_print(*args, sep=" ", end="\n", file=None):
    scrubbed_args = (scrub_filter(str(arg)) for arg in args)
    return unwrapped_print_function(*scrubbed_args, sep=sep, end=end, file=file)


# NOTE: These will wrap the builtin print and the log function, so that there is no way around
#       to log any sensitive data, even if the source of the logging message does not come from
#       a Sensitive object in itself.

Logger._log = scrub_sensitive_data_from_logger
builtins.print = scrub_sensitive_data_from_print
print = scrub_sensitive_data_from_print  # pylint: disable=redefined-builtin

ProtoSensitiveValue.__str__ = wraps(ProtoSensitiveValue.__str__)(
    lambda _: "Sensitive(****)"
)
ProtoSensitiveValue.__repr__ = wraps(ProtoSensitiveValue.__repr__)(
    lambda _: "Sensitive(****)"
)



from bdk.v1.types import book_procedure_descriptor_pb2 as _book_procedure_descriptor_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union
DESCRIPTOR: _descriptor.FileDescriptor

class DiscoverProceduresResponse(_message.Message):
    __slots__ = ('procedures',)
    PROCEDURES_FIELD_NUMBER: _ClassVar[int]
    procedures: _containers.RepeatedCompositeFieldContainer[_book_procedure_descriptor_pb2.BookProcedureDescriptor]

    def __init__(self, procedures: _Optional[_Iterable[_Union[_book_procedure_descriptor_pb2.BookProcedureDescriptor, _Mapping]]]=...) -> None:
        ...


"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_sym_db = _symbol_database.Default()
from ...bdk.v1.requests import request_pb2 as bdk_dot_v1_dot_requests_dot_request__pb2
from ...bdk.v1.responses import response_pb2 as bdk_dot_v1_dot_responses_dot_response__pb2
from ...bdk.v1.services import book_pb2 as bdk_dot_v1_dot_services_dot_book__pb2
from ...bdk.v1.services import library_pb2 as bdk_dot_v1_dot_services_dot_library__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from ...grpc.health.v1 import health_pb2 as grpc_dot_health_dot_v1_dot_health__pb2
from ... import version_pb2 as version__pb2
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0fbdk/v1/v1.proto\x12\x06bdk.v1\x1a\x1dbdk/v1/requests/request.proto\x1a\x1fbdk/v1/responses/response.proto\x1a\x1abdk/v1/services/book.proto\x1a\x1dbdk/v1/services/library.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1bgrpc/health/v1/health.proto\x1a\rversion.protoB\x1fZ\x1dgithub.com/kognitos/bdk-protob\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'bdk.v1.v1_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals['DESCRIPTOR']._options = None
    _globals['DESCRIPTOR']._serialized_options = b'Z\x1dgithub.com/kognitos/bdk-proto'


from bdk.v2.requests import discover_procedures_pb2 as _discover_procedures_pb2
from bdk.v2.requests import environment_information_pb2 as _environment_information_pb2
from bdk.v2.requests import invoke_procedure_pb2 as _invoke_procedure_pb2
from bdk.v2.requests import resolve_promise_pb2 as _resolve_promise_pb2
from bdk.v2.requests import retrieve_book_pb2 as _retrieve_book_pb2
from bdk.v2.requests import retrieve_book_procedures_pb2 as _retrieve_book_procedures_pb2
from bdk.v2.requests import retrieve_books_pb2 as _retrieve_books_pb2
from bdk.v2.requests import retrieve_discoverables_pb2 as _retrieve_discoverables_pb2
from bdk.v2.requests import test_connection_pb2 as _test_connection_pb2
from bdk.v2.responses import discover_procedures_pb2 as _discover_procedures_pb2_1
from bdk.v2.responses import environment_information_pb2 as _environment_information_pb2_1
from bdk.v2.responses import invoke_procedure_pb2 as _invoke_procedure_pb2_1
from bdk.v2.responses import retrieve_book_pb2 as _retrieve_book_pb2_1
from bdk.v2.responses import retrieve_book_procedures_pb2 as _retrieve_book_procedures_pb2_1
from bdk.v2.responses import retrieve_books_pb2 as _retrieve_books_pb2_1
from bdk.v2.responses import retrieve_discoverables_pb2 as _retrieve_discoverables_pb2_1
from bdk.v2.responses import test_connection_pb2 as _test_connection_pb2_1
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar
DESCRIPTOR: _descriptor.FileDescriptor


# pylint: disable=no-name-in-module
from kognitos.bdk.runtime.client.proto.bdk.v1.types.book_procedure_signature_pb2 import \
    BookProcedureSignature as ProtoBookProcedureSignature

from ..book_procedure_signature import BookProcedureSignature
from .noun_phrase import map_noun_phrases


def map_book_procedure_signature(
    data: ProtoBookProcedureSignature,
) -> BookProcedureSignature:
    return BookProcedureSignature(
        english=data.english,
        preposition=data.preposition,
        is_read_only=data.is_read_only,
        verbs=list(data.verbs),
        object=map_noun_phrases(data.object),
        target=map_noun_phrases(data.target),
        outputs=[map_noun_phrases(output) for output in data.outputs],
        proper_nouns=list(data.proper_nouns),
    )



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional
DESCRIPTOR: _descriptor.FileDescriptor

class RetrieveBookProceduresRequest(_message.Message):
    __slots__ = ('name', 'version', 'include_connect')
    NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    INCLUDE_CONNECT_FIELD_NUMBER: _ClassVar[int]
    name: str
    version: str
    include_connect: bool

    def __init__(self, name: _Optional[str]=..., version: _Optional[str]=..., include_connect: bool=...) -> None:
        ...


"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_sym_db = _symbol_database.Default()
from ....bdk.v1.requests import authentication_pb2 as bdk_dot_v1_dot_requests_dot_authentication__pb2
from ....bdk.v1.requests import labels_pb2 as bdk_dot_v1_dot_requests_dot_labels__pb2
from ....bdk.v1.types import config_value_pb2 as bdk_dot_v1_dot_types_dot_config__value__pb2
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n-bdk/v1/types/book_connection_descriptor.proto\x12\x06bdk.v1\x1a$bdk/v1/requests/authentication.proto\x1a\x1cbdk/v1/requests/labels.proto\x1a\x1fbdk/v1/types/config_value.proto"\xf4\x02\n\x18BookConnectionDescriptor\x12\x1b\n\tbook_name\x18\x01 \x01(\tR\x08bookName\x12!\n\x0cbook_version\x18\x02 \x01(\tR\x0bbookVersion\x12#\n\rconnection_id\x18\x03 \x01(\tR\x0cconnectionId\x12>\n\x0eauthentication\x18\x04 \x01(\x0b2\x16.bdk.v1.AuthenticationR\x0eauthentication\x12+\n\x06config\x18\x05 \x03(\x0b2\x13.bdk.v1.ConfigValueR\x06config\x12%\n\x06labels\x18\x06 \x03(\x0b2\r.bdk.v1.LabelR\x06labels\x121\n\x05state\x18\x07 \x01(\x0e2\x1b.bdk.v1.BookConnectionStateR\x05state\x12\x1f\n\x08endpoint\x18\t \x01(\tH\x00R\x08endpoint\x88\x01\x01B\x0b\n\t_endpoint*{\n\x13BookConnectionState\x12!\n\x1dBOOK_CONNECTION_STATE_PENDING\x10\x00\x12\x1f\n\x1bBOOK_CONNECTION_STATE_READY\x10\x01\x12 \n\x1cBOOK_CONNECTION_STATE_FAILED\x10\x02B\x1fZ\x1dgithub.com/kognitos/bdk-protob\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'bdk.v1.types.book_connection_descriptor_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals['DESCRIPTOR']._options = None
    _globals['DESCRIPTOR']._serialized_options = b'Z\x1dgithub.com/kognitos/bdk-proto'
    _globals['_BOOKCONNECTIONSTATE']._serialized_start = 533
    _globals['_BOOKCONNECTIONSTATE']._serialized_end = 656
    _globals['_BOOKCONNECTIONDESCRIPTOR']._serialized_start = 159
    _globals['_BOOKCONNECTIONDESCRIPTOR']._serialized_end = 531


from dataclasses import dataclass, field
from typing import List

from .base import BdkDescriptorBase
from .book_procedure_signature import BookProcedureSignature
from .concept_descriptor import ConceptDescriptor
from .connection_required import ConnectionRequired
from .example_descriptor import ExampleDescriptor
from .question_descriptor import QuestionDescriptor


@dataclass
class BookProcedureDescriptor(BdkDescriptorBase):
    id: str
    short_description: str
    long_description: str
    signature: BookProcedureSignature
    inputs: List[ConceptDescriptor]
    outputs: List[ConceptDescriptor]
    filter_capable: bool
    page_capable: bool
    connection_required: ConnectionRequired
    examples: List[ExampleDescriptor] = field(default_factory=list)
    questions: List[QuestionDescriptor] = field(default_factory=list)



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_sym_db = _symbol_database.Default()
from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rversion.proto\x1a google/protobuf/descriptor.proto:I\n\x10protocol_version\x12\x1c.google.protobuf.FileOptions\x18\xb8\x8e\x03 \x01(\tR\x0fprotocolVersionB)Z\x1dgithub.com/kognitos/bdk-proto\xc2\xf3\x18\x061.32.0b\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'version_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals['DESCRIPTOR']._options = None
    _globals['DESCRIPTOR']._serialized_options = b'Z\x1dgithub.com/kognitos/bdk-proto\xc2\xf3\x18\x061.32.0'


"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_sym_db = _symbol_database.Default()
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15google/api/http.proto\x12\ngoogle.api"y\n\x04Http\x12*\n\x05rules\x18\x01 \x03(\x0b2\x14.google.api.HttpRuleR\x05rules\x12E\n\x1ffully_decode_reserved_expansion\x18\x02 \x01(\x08R\x1cfullyDecodeReservedExpansion"\xda\x02\n\x08HttpRule\x12\x1a\n\x08selector\x18\x01 \x01(\tR\x08selector\x12\x12\n\x03get\x18\x02 \x01(\tH\x00R\x03get\x12\x12\n\x03put\x18\x03 \x01(\tH\x00R\x03put\x12\x14\n\x04post\x18\x04 \x01(\tH\x00R\x04post\x12\x18\n\x06delete\x18\x05 \x01(\tH\x00R\x06delete\x12\x16\n\x05patch\x18\x06 \x01(\tH\x00R\x05patch\x127\n\x06custom\x18\x08 \x01(\x0b2\x1d.google.api.CustomHttpPatternH\x00R\x06custom\x12\x12\n\x04body\x18\x07 \x01(\tR\x04body\x12#\n\rresponse_body\x18\x0c \x01(\tR\x0cresponseBody\x12E\n\x13additional_bindings\x18\x0b \x03(\x0b2\x14.google.api.HttpRuleR\x12additionalBindingsB\t\n\x07pattern";\n\x11CustomHttpPattern\x12\x12\n\x04kind\x18\x01 \x01(\tR\x04kind\x12\x12\n\x04path\x18\x02 \x01(\tR\x04pathBg\n\x0ecom.google.apiB\tHttpProtoP\x01ZAgoogle.golang.org/genproto/googleapis/api/annotations;annotations\xa2\x02\x04GAPIb\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'google.api.http_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals['DESCRIPTOR']._options = None
    _globals['DESCRIPTOR']._serialized_options = b'\n\x0ecom.google.apiB\tHttpProtoP\x01ZAgoogle.golang.org/genproto/googleapis/api/annotations;annotations\xa2\x02\x04GAPI'
    _globals['_HTTP']._serialized_start = 37
    _globals['_HTTP']._serialized_end = 158
    _globals['_HTTPRULE']._serialized_start = 161
    _globals['_HTTPRULE']._serialized_end = 507
    _globals['_CUSTOMHTTPPATTERN']._serialized_start = 509
    _globals['_CUSTOMHTTPPATTERN']._serialized_end = 568


from bdk.v2.types import book_procedure_signature_pb2 as _book_procedure_signature_pb2
from bdk.v2.types import concept_descriptor_pb2 as _concept_descriptor_pb2
from bdk.v2.types import connection_required_pb2 as _connection_required_pb2
from bdk.v2.types import example_descriptor_pb2 as _example_descriptor_pb2
from bdk.v2.types import question_descriptor_pb2 as _question_descriptor_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union
DESCRIPTOR: _descriptor.FileDescriptor

class BookProcedureDescriptor(_message.Message):
    __slots__ = ('id', 'signature', 'inputs', 'outputs', 'short_description', 'long_description', 'filter_capable', 'page_capable', 'connection_required', 'is_discovered', 'questions', 'examples', 'is_async', 'connection_requirement_level', 'is_mutation')
    ID_FIELD_NUMBER: _ClassVar[int]
    SIGNATURE_FIELD_NUMBER: _ClassVar[int]
    INPUTS_FIELD_NUMBER: _ClassVar[int]
    OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    SHORT_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    LONG_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    FILTER_CAPABLE_FIELD_NUMBER: _ClassVar[int]
    PAGE_CAPABLE_FIELD_NUMBER: _ClassVar[int]
    CONNECTION_REQUIRED_FIELD_NUMBER: _ClassVar[int]
    IS_DISCOVERED_FIELD_NUMBER: _ClassVar[int]
    QUESTIONS_FIELD_NUMBER: _ClassVar[int]
    EXAMPLES_FIELD_NUMBER: _ClassVar[int]
    IS_ASYNC_FIELD_NUMBER: _ClassVar[int]
    CONNECTION_REQUIREMENT_LEVEL_FIELD_NUMBER: _ClassVar[int]
    IS_MUTATION_FIELD_NUMBER: _ClassVar[int]
    id: str
    signature: _book_procedure_signature_pb2.BookProcedureSignature
    inputs: _containers.RepeatedCompositeFieldContainer[_concept_descriptor_pb2.ConceptDescriptor]
    outputs: _containers.RepeatedCompositeFieldContainer[_concept_descriptor_pb2.ConceptDescriptor]
    short_description: str
    long_description: str
    filter_capable: bool
    page_capable: bool
    connection_required: bool
    is_discovered: bool
    questions: _containers.RepeatedCompositeFieldContainer[_question_descriptor_pb2.QuestionDescriptor]
    examples: _containers.RepeatedCompositeFieldContainer[_example_descriptor_pb2.ExampleDescriptor]
    is_async: bool
    connection_requirement_level: _connection_required_pb2.ConnectionRequired
    is_mutation: bool

    def __init__(self, id: _Optional[str]=..., signature: _Optional[_Union[_book_procedure_signature_pb2.BookProcedureSignature, _Mapping]]=..., inputs: _Optional[_Iterable[_Union[_concept_descriptor_pb2.ConceptDescriptor, _Mapping]]]=..., outputs: _Optional[_Iterable[_Union[_concept_descriptor_pb2.ConceptDescriptor, _Mapping]]]=..., short_description: _Optional[str]=..., long_description: _Optional[str]=..., filter_capable: bool=..., page_capable: bool=..., connection_required: bool=..., is_discovered: bool=..., questions: _Optional[_Iterable[_Union[_question_descriptor_pb2.QuestionDescriptor, _Mapping]]]=..., examples: _Optional[_Iterable[_Union[_example_descriptor_pb2.ExampleDescriptor, _Mapping]]]=..., is_async: bool=..., connection_requirement_level: _Optional[_Union[_connection_required_pb2.ConnectionRequired, str]]=..., is_mutation: bool=...) -> None:
        ...

class BookProcedureDescriptorV2(_message.Message):
    __slots__ = ('id', 'signature', 'inputs', 'outputs', 'short_description', 'long_description', 'filter_capable', 'page_capable', 'connection_required', 'is_discovered', 'questions', 'examples', 'is_async', 'is_mutation')
    ID_FIELD_NUMBER: _ClassVar[int]
    SIGNATURE_FIELD_NUMBER: _ClassVar[int]
    INPUTS_FIELD_NUMBER: _ClassVar[int]
    OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    SHORT_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    LONG_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    FILTER_CAPABLE_FIELD_NUMBER: _ClassVar[int]
    PAGE_CAPABLE_FIELD_NUMBER: _ClassVar[int]
    CONNECTION_REQUIRED_FIELD_NUMBER: _ClassVar[int]
    IS_DISCOVERED_FIELD_NUMBER: _ClassVar[int]
    QUESTIONS_FIELD_NUMBER: _ClassVar[int]
    EXAMPLES_FIELD_NUMBER: _ClassVar[int]
    IS_ASYNC_FIELD_NUMBER: _ClassVar[int]
    IS_MUTATION_FIELD_NUMBER: _ClassVar[int]
    id: str
    signature: _book_procedure_signature_pb2.BookProcedureSignature
    inputs: _containers.RepeatedCompositeFieldContainer[_concept_descriptor_pb2.ConceptDescriptor]
    outputs: _containers.RepeatedCompositeFieldContainer[_concept_descriptor_pb2.ConceptDescriptor]
    short_description: str
    long_description: str
    filter_capable: bool
    page_capable: bool
    connection_required: _connection_required_pb2.ConnectionRequired
    is_discovered: bool
    questions: _containers.RepeatedCompositeFieldContainer[_question_descriptor_pb2.QuestionDescriptor]
    examples: _containers.RepeatedCompositeFieldContainer[_example_descriptor_pb2.ExampleDescriptor]
    is_async: bool
    is_mutation: bool

    def __init__(self, id: _Optional[str]=..., signature: _Optional[_Union[_book_procedure_signature_pb2.BookProcedureSignature, _Mapping]]=..., inputs: _Optional[_Iterable[_Union[_concept_descriptor_pb2.ConceptDescriptor, _Mapping]]]=..., outputs: _Optional[_Iterable[_Union[_concept_descriptor_pb2.ConceptDescriptor, _Mapping]]]=..., short_description: _Optional[str]=..., long_description: _Optional[str]=..., filter_capable: bool=..., page_capable: bool=..., connection_required: _Optional[_Union[_connection_required_pb2.ConnectionRequired, str]]=..., is_discovered: bool=..., questions: _Optional[_Iterable[_Union[_question_descriptor_pb2.QuestionDescriptor, _Mapping]]]=..., examples: _Optional[_Iterable[_Union[_example_descriptor_pb2.ExampleDescriptor, _Mapping]]]=..., is_async: bool=..., is_mutation: bool=...) -> None:
        ...


from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional
DESCRIPTOR: _descriptor.FileDescriptor

class RetrieveBooksRequest(_message.Message):
    __slots__ = ('page_size', 'page_token')
    PAGE_SIZE_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    page_size: int
    page_token: str

    def __init__(self, page_size: _Optional[int]=..., page_token: _Optional[str]=...) -> None:
        ...


from typing import Iterable, List

from ..input_concept import InputConcept
from ..proto.bdk.v1.types.concept_value_pb2 import \
    ConceptValue  # pylint: disable=no-name-in-module
from ..proto.bdk.v1.types.noun_phrase_pb2 import (  # pylint: disable=no-name-in-module
    NounPhrase, NounPhrases)
from . import value as value_mapper


def _map_concept(concept: InputConcept) -> ConceptValue:
    noun_phrases = concept.noun_phrases.noun_phrases
    value = concept.value

    return ConceptValue(
        noun_phrases=NounPhrases(
            noun_phrases=[
                NounPhrase(head=np.head, modifiers=np.modifiers) for np in noun_phrases
            ]
        ),
        value=value_mapper.map_value(value),
    )


def map_input_concepts(
    input_concepts: List[InputConcept],
) -> Iterable[ConceptValue]:
    return list(map(_map_concept, input_concepts))



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar
DESCRIPTOR: _descriptor.FileDescriptor

class CredentialType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    CredentialTypeText: _ClassVar[CredentialType]
    CredentialTypeSensitiveText: _ClassVar[CredentialType]
CredentialTypeText: CredentialType
CredentialTypeSensitiveText: CredentialType


from bdk.v1.requests import request_pb2 as _request_pb2
from bdk.v1.responses import response_pb2 as _response_pb2
from bdk.v1.services import book_pb2 as _book_pb2
from bdk.v1.services import library_pb2 as _library_pb2
from google.protobuf import struct_pb2 as _struct_pb2
from grpc.health.v1 import health_pb2 as _health_pb2
import version_pb2 as _version_pb2
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar
DESCRIPTOR: _descriptor.FileDescriptor


from bdk.v1.types import promise_pb2 as _promise_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union
DESCRIPTOR: _descriptor.FileDescriptor

class PromiseResponse(_message.Message):
    __slots__ = ('promise',)
    PROMISE_FIELD_NUMBER: _ClassVar[int]
    promise: _promise_pb2.Promise

    def __init__(self, promise: _Optional[_Union[_promise_pb2.Promise, _Mapping]]=...) -> None:
        ...


import base64
import time
from functools import wraps
from typing import Optional

import boto3
from botocore.config import Config
from botocore.exceptions import ClientError

from kognitos.bdk.runtime.client.base_transport import BaseTransport

from .proto.bdk.v1.requests.request_pb2 import \
    Request  # pylint: disable=no-name-in-module
from .proto.bdk.v1.responses.response_pb2 import \
    Response  # pylint: disable=no-name-in-module
from .transport import TransportError

# Extend timeouts & enable keep-alive
DEFAULT_CONFIG = Config(
    retries={"max_attempts": 0},  # No automatic retries
    connect_timeout=900,  # Max Lambda runtime (15 min)
    read_timeout=900,  # Wait up to 15 min for response
)


class RetryLimitExceededException(Exception):
    """Exception raised when the maximum number of retries is exceeded."""


def retry(backoff_factor, max_retries):
    def decorator_retry(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            retries = 0
            while retries < max_retries:
                try:
                    return func(*args, **kwargs)
                except ClientError as e:
                    if "lambda is initializing your function" in str(e).lower():
                        sleep_time = backoff_factor**retries
                        time.sleep(sleep_time)
                        retries += 1
                    else:
                        raise
            raise RetryLimitExceededException(
                f"Failed to invoke function after {max_retries} retries"
            )

        return wrapper

    return decorator_retry


class LambdaTransport(BaseTransport):
    def __init__(
        self,
        function_name: str,
        sleep_retry_time: float = 10.0,
        config: Optional[Config] = None,
    ):
        self.lambda_client = boto3.client("lambda", config=config or DEFAULT_CONFIG)
        self.function_name = function_name
        self.sleep_retry_time = sleep_retry_time

    def get_url(self) -> str:
        return self.function_name

    def send(self, request: Request) -> Response:
        invoke_function = retry(backoff_factor=self.sleep_retry_time, max_retries=3)(
            self._invoke_function
        )
        return invoke_function(request)

    def _invoke_function(self, request: Request) -> Response:
        response = self.lambda_client.invoke(
            FunctionName=self.function_name,
            Payload=f'"{base64.b64encode(request.SerializeToString()).decode("utf-8")}"',
        )

        try:
            raw_response = response["Payload"].read().decode("utf-8").replace('"', "")

            response_payload = base64.b64decode(raw_response, validate=True)

            return Response.FromString(response_payload)
        except Exception as e:
            if "response_payload" in locals():
                response = f"response_payload: {locals().get('response_payload')}"
            elif "raw_response" in locals():
                response = f"raw_response: {locals().get('raw_response')}"
            raise TransportError(
                f"Could not decode the response <<<{response}>>> got by calling(LambdaTransport) <<<{self.function_name}>>> with the request <<<{request.WhichOneof('request_discriminator')}>>>"
            ) from e



from kognitos.bdk.runtime.client.book_authentication_descriptor import (
    BookAuthenticationDescriptor, BookCustomAuthenticationDescriptor,
    BookOAuthAuthenticationDescriptor, OauthFlow, OauthProvider)
# pylint: disable=no-name-in-module
from kognitos.bdk.runtime.client.mappers.noun_phrase import map_noun_phrase
from kognitos.bdk.runtime.client.proto.bdk.v1.types.book_authentication_descriptor_pb2 import \
    BookAuthenticationDescriptor as ProtoBookAuthenticationDescriptor

from . import credential_descriptor, oauth_argument_descriptor


def map_book_authentication_descriptor(
    data: ProtoBookAuthenticationDescriptor,
) -> BookAuthenticationDescriptor:
    auth_type = data.WhichOneof("authentication_discriminator")

    if auth_type == "oauth":
        auth = data.oauth
        return BookOAuthAuthenticationDescriptor(
            id=auth.id,
            provider=OauthProvider(auth.provider),
            flows=[OauthFlow(f) for f in auth.flows],
            authorize_endpoint=auth.authorize_endpoint,
            token_endpoint=auth.token_endpoint,
            scope=auth.scope,  # type: ignore
            name=auth.name or "oauth",
            arguments=[
                oauth_argument_descriptor.map_oauth_argument_descriptor(arg)
                for arg in auth.arguments
            ],
        )
    if auth_type == "custom":
        auth = data.custom
        return BookCustomAuthenticationDescriptor(
            id=auth.id,
            credentials=[
                credential_descriptor.map_credential_descriptor(cd)
                for cd in auth.credentials
            ],
            description=auth.description,
            name=auth.name or auth.id,
            noun_phrase=map_noun_phrase(auth.noun_phrase),
        )

    raise ValueError(f"Cannot map authentication of type {auth_type}")



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_sym_db = _symbol_database.Default()
from ....bdk.v2.types import oauth_argument_descriptor_pb2 as bdk_dot_v2_dot_types_dot_oauth__argument__descriptor__pb2
from ....bdk.v2.types import oauth_flow_pb2 as bdk_dot_v2_dot_types_dot_oauth__flow__pb2
from ....bdk.v2.types import oauth_provider_pb2 as bdk_dot_v2_dot_types_dot_oauth__provider__pb2
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n6bdk/v2/types/book_oauh_authentication_descriptor.proto\x12\x06bdk.v2\x1a,bdk/v2/types/oauth_argument_descriptor.proto\x1a\x1dbdk/v2/types/oauth_flow.proto\x1a!bdk/v2/types/oauth_provider.proto"\xd2\x02\n!BookOAuthAuthenticationDescriptor\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x125\n\x08provider\x18\x02 \x01(\x0e2\x15.bdk.v2.OAuthProviderB\x02\x18\x01R\x08provider\x12\'\n\x05flows\x18\x03 \x03(\x0e2\x11.bdk.v2.OAuthFlowR\x05flows\x12-\n\x12authorize_endpoint\x18\x04 \x01(\tR\x11authorizeEndpoint\x12%\n\x0etoken_endpoint\x18\x05 \x01(\tR\rtokenEndpoint\x12\x14\n\x05scope\x18\x06 \x03(\tR\x05scope\x12\x12\n\x04name\x18\x07 \x01(\tR\x04name\x12=\n\targuments\x18\x08 \x03(\x0b2\x1f.bdk.v2.OAuthArgumentDescriptorR\targumentsB\x1fZ\x1dgithub.com/kognitos/bdk-protob\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'bdk.v2.types.book_oauh_authentication_descriptor_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals['DESCRIPTOR']._options = None
    _globals['DESCRIPTOR']._serialized_options = b'Z\x1dgithub.com/kognitos/bdk-proto'
    _globals['_BOOKOAUTHAUTHENTICATIONDESCRIPTOR'].fields_by_name['provider']._options = None
    _globals['_BOOKOAUTHAUTHENTICATIONDESCRIPTOR'].fields_by_name['provider']._serialized_options = b'\x18\x01'
    _globals['_BOOKOAUTHAUTHENTICATIONDESCRIPTOR']._serialized_start = 179
    _globals['_BOOKOAUTHAUTHENTICATIONDESCRIPTOR']._serialized_end = 517


from bdk.v2.requests import context_pb2 as _context_pb2
from bdk.v2.requests import discover_procedures_pb2 as _discover_procedures_pb2
from bdk.v2.requests import environment_information_pb2 as _environment_information_pb2
from bdk.v2.requests import invoke_procedure_pb2 as _invoke_procedure_pb2
from bdk.v2.requests import resolve_promise_pb2 as _resolve_promise_pb2
from bdk.v2.requests import retrieve_book_pb2 as _retrieve_book_pb2
from bdk.v2.requests import retrieve_book_procedures_pb2 as _retrieve_book_procedures_pb2
from bdk.v2.requests import retrieve_books_pb2 as _retrieve_books_pb2
from bdk.v2.requests import retrieve_discoverables_pb2 as _retrieve_discoverables_pb2
from bdk.v2.requests import retrieve_tags_pb2 as _retrieve_tags_pb2
from bdk.v2.requests import test_connection_pb2 as _test_connection_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union
DESCRIPTOR: _descriptor.FileDescriptor

class Request(_message.Message):
    __slots__ = ('context', 'environment_information', 'retrieve_books', 'retrieve_book', 'retrieve_procedures', 'test_connection', 'invoke_procedure', 'discover_procedures', 'retrieve_discoverables', 'retrieve_tags', 'resolve_promise')
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_INFORMATION_FIELD_NUMBER: _ClassVar[int]
    RETRIEVE_BOOKS_FIELD_NUMBER: _ClassVar[int]
    RETRIEVE_BOOK_FIELD_NUMBER: _ClassVar[int]
    RETRIEVE_PROCEDURES_FIELD_NUMBER: _ClassVar[int]
    TEST_CONNECTION_FIELD_NUMBER: _ClassVar[int]
    INVOKE_PROCEDURE_FIELD_NUMBER: _ClassVar[int]
    DISCOVER_PROCEDURES_FIELD_NUMBER: _ClassVar[int]
    RETRIEVE_DISCOVERABLES_FIELD_NUMBER: _ClassVar[int]
    RETRIEVE_TAGS_FIELD_NUMBER: _ClassVar[int]
    RESOLVE_PROMISE_FIELD_NUMBER: _ClassVar[int]
    context: _context_pb2.Context
    environment_information: _environment_information_pb2.EnvironmentInformationRequest
    retrieve_books: _retrieve_books_pb2.RetrieveBooksRequest
    retrieve_book: _retrieve_book_pb2.RetrieveBookRequest
    retrieve_procedures: _retrieve_book_procedures_pb2.RetrieveBookProceduresRequest
    test_connection: _test_connection_pb2.TestConnectionRequest
    invoke_procedure: _invoke_procedure_pb2.InvokeProcedureRequest
    discover_procedures: _discover_procedures_pb2.DiscoverProceduresRequest
    retrieve_discoverables: _retrieve_discoverables_pb2.RetrieveDiscoverablesRequest
    retrieve_tags: _retrieve_tags_pb2.RetrieveTagsRequest
    resolve_promise: _resolve_promise_pb2.ResolvePromiseRequest

    def __init__(self, context: _Optional[_Union[_context_pb2.Context, _Mapping]]=..., environment_information: _Optional[_Union[_environment_information_pb2.EnvironmentInformationRequest, _Mapping]]=..., retrieve_books: _Optional[_Union[_retrieve_books_pb2.RetrieveBooksRequest, _Mapping]]=..., retrieve_book: _Optional[_Union[_retrieve_book_pb2.RetrieveBookRequest, _Mapping]]=..., retrieve_procedures: _Optional[_Union[_retrieve_book_procedures_pb2.RetrieveBookProceduresRequest, _Mapping]]=..., test_connection: _Optional[_Union[_test_connection_pb2.TestConnectionRequest, _Mapping]]=..., invoke_procedure: _Optional[_Union[_invoke_procedure_pb2.InvokeProcedureRequest, _Mapping]]=..., discover_procedures: _Optional[_Union[_discover_procedures_pb2.DiscoverProceduresRequest, _Mapping]]=..., retrieve_discoverables: _Optional[_Union[_retrieve_discoverables_pb2.RetrieveDiscoverablesRequest, _Mapping]]=..., retrieve_tags: _Optional[_Union[_retrieve_tags_pb2.RetrieveTagsRequest, _Mapping]]=..., resolve_promise: _Optional[_Union[_resolve_promise_pb2.ResolvePromiseRequest, _Mapping]]=...) -> None:
        ...


"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_sym_db = _symbol_database.Default()
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dbdk/v2/requests/context.proto\x12\x06bdk.v2"\xaf\x02\n\x07Context\x12\x1e\n\x08trace_id\x18\x01 \x01(\x04H\x00R\x07traceId\x88\x01\x01\x12\x1c\n\x07span_id\x18\x02 \x01(\x04H\x01R\x06spanId\x88\x01\x01\x12 \n\tworker_id\x18\x03 \x01(\tH\x02R\x08workerId\x88\x01\x01\x12(\n\rdepartment_id\x18\x04 \x01(\tH\x03R\x0cdepartmentId\x88\x01\x01\x12&\n\x0cknowledge_id\x18\x05 \x01(\tH\x04R\x0bknowledgeId\x88\x01\x01\x12\x1c\n\x07line_id\x18\x06 \x01(\tH\x05R\x06lineId\x88\x01\x01B\x0b\n\t_trace_idB\n\n\x08_span_idB\x0c\n\n_worker_idB\x10\n\x0e_department_idB\x0f\n\r_knowledge_idB\n\n\x08_line_idB\x1fZ\x1dgithub.com/kognitos/bdk-protob\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'bdk.v2.requests.context_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals['DESCRIPTOR']._options = None
    _globals['DESCRIPTOR']._serialized_options = b'Z\x1dgithub.com/kognitos/bdk-proto'
    _globals['_CONTEXT']._serialized_start = 42
    _globals['_CONTEXT']._serialized_end = 345


from .client import Client
from .http_transport import HTTPTransport
from .lambda_rie_transport import LambdaRIETransport
from .lambda_transport import LambdaTransport
from .transport import Transport



from kognitos.bdk.runtime.client.promise import Promise
from kognitos.bdk.runtime.client.proto.bdk.v1.types.promise_pb2 import \
    Promise as ProtoPromise  # pylint: disable=no-name-in-module

from . import value as value_mapper


def map_promise(promise: ProtoPromise) -> Promise:
    return Promise(
        promise_resolver_function_name=promise.promise_resolver_function_name,
        data=value_mapper.map_value(promise.data),
    )



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union
DESCRIPTOR: _descriptor.FileDescriptor

class NounPhrase(_message.Message):
    __slots__ = ('modifiers', 'head')
    MODIFIERS_FIELD_NUMBER: _ClassVar[int]
    HEAD_FIELD_NUMBER: _ClassVar[int]
    modifiers: _containers.RepeatedScalarFieldContainer[str]
    head: str

    def __init__(self, modifiers: _Optional[_Iterable[str]]=..., head: _Optional[str]=...) -> None:
        ...

class NounPhrases(_message.Message):
    __slots__ = ('noun_phrases',)
    NOUN_PHRASES_FIELD_NUMBER: _ClassVar[int]
    noun_phrases: _containers.RepeatedCompositeFieldContainer[NounPhrase]

    def __init__(self, noun_phrases: _Optional[_Iterable[_Union[NounPhrase, _Mapping]]]=...) -> None:
        ...


# pylint: disable=no-name-in-module
from kognitos.bdk.runtime.client.proto.bdk.v1.responses.environment_information_pb2 import \
    EnvironmentInformationResponse as ProtoEnvironmentInformation

from ..environment_information import EnvironmentInformation


def map_environment_information(
    data: ProtoEnvironmentInformation,
) -> EnvironmentInformation:
    return EnvironmentInformation(
        version=data.version,
        runtime_name=data.runtime_name,
        runtime_version=data.runtime_version,
        bci_protocol_version=data.bci_protocol_version,
        api_version=data.api_version,
        path=list(data.path),
    )



from abc import abstractmethod
from typing import Any, List, Literal, Optional, Tuple

from . import input_mappers, mappers
from .book_descriptor import BookDescriptor
from .book_procedure_descriptor import BookProcedureDescriptor
from .discoverable import Discoverable
from .environment_information import EnvironmentInformation
from .exceptions import UnexpectedResponse, handle_exception
from .expression import Expression
from .input_mappers.input_concepts import InputConcept
from .offload import Offload
from .promise import Promise
from .proto.bdk.v1.requests.context_pb2 import \
    Context  # pylint: disable=no-name-in-module
from .proto.bdk.v1.requests.discover_procedures_pb2 import \
    DiscoverProceduresRequest  # pylint: disable=no-name-in-module
from .proto.bdk.v1.requests.environment_information_pb2 import \
    EnvironmentInformationRequest  # pylint: disable=no-name-in-module
from .proto.bdk.v1.requests.invoke_procedure_pb2 import \
    InvokeProcedureRequest  # pylint: disable=no-name-in-module
from .proto.bdk.v1.requests.request_pb2 import \
    Request  # pylint: disable=no-name-in-module
from .proto.bdk.v1.requests.resolve_promise_pb2 import \
    ResolvePromiseRequest  # pylint: disable=no-name-in-module
from .proto.bdk.v1.requests.retrieve_book_pb2 import \
    RetrieveBookRequest  # pylint: disable=no-name-in-module
from .proto.bdk.v1.requests.retrieve_book_procedures_pb2 import \
    RetrieveBookProceduresRequest  # pylint: disable=no-name-in-module
from .proto.bdk.v1.requests.retrieve_books_pb2 import \
    RetrieveBooksRequest  # pylint: disable=no-name-in-module
from .proto.bdk.v1.requests.retrieve_discoverables_pb2 import \
    RetrieveDiscoverablesRequest  # pylint: disable=no-name-in-module
from .proto.bdk.v1.requests.retrieve_tags_pb2 import \
    RetrieveTagsRequest  # pylint: disable=no-name-in-module
from .proto.bdk.v1.requests.test_connection_pb2 import \
    TestConnectionRequest  # pylint: disable=no-name-in-module
from .proto.bdk.v1.responses.response_pb2 import \
    Response  # pylint: disable=no-name-in-module
from .question_answer import AnsweredQuestion
from .request_context import RequestContext
from .transport import InvokeProcedureResponse, Transport


def map_context(context: Optional[RequestContext]) -> Context:
    if not context:
        context = RequestContext()

    return Context(
        trace_id=context.trace_id,
        span_id=context.span_id,
        worker_id=context.worker_id,
        department_id=context.department_id,
        knowledge_id=context.knowledge_id,
        line_id=context.line_id,
    )


ResponseFieldLiteral = Literal[
    "environment_information",
    "error",
    "invoked_procedure",
    "retrieved_book",
    "retrieved_books",
    "retrieved_procedures",
    "tested_connection",
    "discovered_procedures",
    "promise",
    "question",
    "retrieved_discoverables",
    "retrieved_tags",
]


def handle_response(
    response: Response,
    response_field: ResponseFieldLiteral | List[ResponseFieldLiteral],
):
    if not isinstance(response_field, list):
        response_field = [response_field]

    if response.HasField("error"):
        handle_exception(response.error)
    if not any(response.HasField(field) for field in response_field):
        raise UnexpectedResponse(
            f"Didn't get any of the expected responses {response_field} from server"
        )
    response_field = next(field for field in response_field if response.HasField(field))
    return getattr(response, response_field)


class BaseTransport(Transport):
    @abstractmethod
    def send(self, request: Request) -> Response:
        pass

    def retrieve_books(
        self, context: Optional[RequestContext] = None
    ) -> List[BookDescriptor]:
        request = Request(
            retrieve_books=RetrieveBooksRequest(), context=map_context(context)
        )
        response = handle_response(self.send(request), "retrieved_books")
        return [mappers.map_book_descriptor(b) for b in response.books]

    def environment_information(
        self, context: Optional[RequestContext] = None
    ) -> EnvironmentInformation:
        request = Request(
            environment_information=EnvironmentInformationRequest(),
            context=map_context(context),
        )
        return mappers.map_environment_information(
            handle_response(self.send(request), "environment_information")
        )

    def retrieve_book(
        self, name, version, context: Optional[RequestContext] = None
    ) -> BookDescriptor:
        request = Request(
            retrieve_book=RetrieveBookRequest(name=name, version=version),
            context=map_context(context),
        )
        response = handle_response(self.send(request), "retrieved_book")
        return mappers.map_book_descriptor(response.book)

    def retrieve_procedures(
        self,
        name: str,
        version: str,
        include_connect: bool,
        context: Optional[RequestContext] = None,
    ) -> List[BookProcedureDescriptor]:
        request = Request(
            retrieve_procedures=RetrieveBookProceduresRequest(
                name=name, version=version, include_connect=include_connect
            ),
            context=map_context(context),
        )

        response = handle_response(self.send(request), "retrieved_procedures")

        return [mappers.map_book_procedure_descriptor(ps) for ps in response.procedures]

    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def test_connection(
        self,
        name: str,
        version: str,
        authentication_id: Optional[str],
        authentication_credentials: Optional[List[Tuple[str, Any]]],
        context: Optional[RequestContext] = None,
    ) -> Any:
        authentication = input_mappers.map_authentication(
            authentication_id, authentication_credentials
        )

        request = Request(
            test_connection=TestConnectionRequest(
                name=name, version=version, authentication=authentication
            ),
            context=map_context(context),
        )

        response = handle_response(self.send(request), "tested_connection")

        return mappers.map_test_connection_response(response)

    # pylint: disable=too-many-arguments, too-many-locals
    def invoke_procedure(
        self,
        name: str,
        version: str,
        procedure_id: str,
        authentication_id: Optional[str],
        authentication_credentials: Optional[List[Tuple[str, Any]]],
        input_concepts: List[InputConcept],
        context: Optional[RequestContext] = None,
        filter_expression: Optional[Expression] = None,
        offload: Optional[Offload] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        answered_questions: Optional[List[AnsweredQuestion]] = None,
    ) -> InvokeProcedureResponse:
        if answered_questions is None:
            answered_questions = []

        request = Request(
            invoke_procedure=InvokeProcedureRequest(
                name=name,
                version=version,
                procedure_id=procedure_id,
                authentication=input_mappers.map_authentication(
                    authentication_id, authentication_credentials
                ),
                input_concepts=input_mappers.map_input_concepts(input_concepts),
                filter_expression=input_mappers.map_expression(filter_expression),
                offload=input_mappers.map_offload(offload),
                offset=offset,
                limit=limit,
                answered_questions=input_mappers.map_answered_questions(
                    answered_questions
                ),
            ),
            context=map_context(context),
        )

        response = handle_response(
            self.send(request), ["invoked_procedure", "question", "promise"]
        )

        if hasattr(
            response, "questions"
        ):  # HasField raises ValueError instead of yielding False
            return mappers.map_questions(response.questions)

        if hasattr(
            response, "promise"
        ):  # HasField raises ValueError instead of yielding False
            return mappers.map_promise(response.promise)

        return mappers.map_concept_values(response.output_concepts)

    def discover_procedures(
        self,
        name: str,
        version: str,
        authentication_id: Optional[str],
        authentication_credentials: Optional[List[Tuple[str, Any]]],
        what: List[str],
        context: Optional[RequestContext] = None,
    ) -> List[BookProcedureDescriptor]:

        request = Request(
            discover_procedures=DiscoverProceduresRequest(
                name=name,
                version=version,
                authentication=input_mappers.map_authentication(
                    authentication_id, authentication_credentials
                ),
                what=what,
            ),
            context=map_context(context),
        )

        response = handle_response(self.send(request), "discovered_procedures")

        return [mappers.map_book_procedure_descriptor(p) for p in response.procedures]

    def retrieve_discoverables(
        self,
        name: str,
        version: str,
        authentication_id: Optional[str],
        authentication_credentials: Optional[List[Tuple[str, Any]]],
        search: Optional[str],
        limit: Optional[int],
        offset: Optional[int],
        context: Optional[RequestContext] = None,
    ) -> List[Discoverable]:

        request = Request(
            retrieve_discoverables=RetrieveDiscoverablesRequest(
                name=name,
                version=version,
                authentication=input_mappers.map_authentication(
                    authentication_id, authentication_credentials
                ),
                search=search,
                limit=limit,
                offset=offset,
            ),
            context=map_context(context),
        )

        response = handle_response(self.send(request), "retrieved_discoverables")

        return mappers.map_discoverables(response)

    def retrieve_tags(self, context: Optional[RequestContext] = None) -> List[str]:
        request = Request(
            retrieve_tags=RetrieveTagsRequest(),
            context=map_context(context),
        )
        response = handle_response(self.send(request), "retrieved_tags")
        return response.tags

    def resolve_promise(
        self,
        name: str,
        version: str,
        procedure_id: str,
        authentication_id: Optional[str],
        authentication_credentials: Optional[List[Tuple[str, Any]]],
        promise: Promise,
        configurations: List[InputConcept],
        offload: Optional[Offload] = None,
        answered_questions: Optional[List[AnsweredQuestion]] = None,
        context: Optional[RequestContext] = None,
    ) -> InvokeProcedureResponse:
        if answered_questions is None:
            answered_questions = []

        request = Request(
            resolve_promise=ResolvePromiseRequest(
                name=name,
                version=version,
                authentication=input_mappers.map_authentication(
                    authentication_id, authentication_credentials
                ),
                procedure_id=procedure_id,
                promise=input_mappers.map_promise(promise),
                offload=input_mappers.map_offload(offload),
                answered_questions=input_mappers.map_answered_questions(
                    answered_questions
                ),
                configurations=input_mappers.map_input_concepts(configurations),
            ),
            context=map_context(context),
        )

        response = handle_response(
            self.send(request), ["invoked_procedure", "question", "promise"]
        )

        if hasattr(
            response, "questions"
        ):  # HasField raises ValueError instead of yielding False
            return mappers.map_questions(response.questions)

        if hasattr(
            response, "promise"
        ):  # HasField raises ValueError instead of yielding False
            return mappers.map_promise(response.promise)

        return mappers.map_concept_values(response.output_concepts)



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional
DESCRIPTOR: _descriptor.FileDescriptor

class Date(_message.Message):
    __slots__ = ('year', 'month', 'day')
    YEAR_FIELD_NUMBER: _ClassVar[int]
    MONTH_FIELD_NUMBER: _ClassVar[int]
    DAY_FIELD_NUMBER: _ClassVar[int]
    year: int
    month: int
    day: int

    def __init__(self, year: _Optional[int]=..., month: _Optional[int]=..., day: _Optional[int]=...) -> None:
        ...


from typing import Iterable, List

from kognitos.bdk.runtime.client.question_answer import AnsweredQuestion

# pylint: disable=no-name-in-module
from ..proto.bdk.v1.types.answered_question_pb2 import \
    AnsweredQuestion as ProtoAnsweredQuestion
from .noun_phrase import map_noun_phrases
from .value import map_value


def _map_answered_question(
    answered_question: AnsweredQuestion,
) -> ProtoAnsweredQuestion:
    return ProtoAnsweredQuestion(
        noun_phrases=map_noun_phrases(answered_question.noun_phrases),
        answer=map_value(answered_question.answer),
    )


def map_answered_questions(
    answered_questions: List[AnsweredQuestion],
) -> Iterable[ProtoAnsweredQuestion]:
    return list(map(_map_answered_question, answered_questions))



from bdk.v1.types import noun_phrase_pb2 as _noun_phrase_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union
DESCRIPTOR: _descriptor.FileDescriptor

class BookProcedureSignature(_message.Message):
    __slots__ = ('english', 'verbs', 'object', 'preposition', 'target', 'outputs', 'is_read_only', 'proper_nouns')
    ENGLISH_FIELD_NUMBER: _ClassVar[int]
    VERBS_FIELD_NUMBER: _ClassVar[int]
    OBJECT_FIELD_NUMBER: _ClassVar[int]
    PREPOSITION_FIELD_NUMBER: _ClassVar[int]
    TARGET_FIELD_NUMBER: _ClassVar[int]
    OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    IS_READ_ONLY_FIELD_NUMBER: _ClassVar[int]
    PROPER_NOUNS_FIELD_NUMBER: _ClassVar[int]
    english: str
    verbs: _containers.RepeatedScalarFieldContainer[str]
    object: _noun_phrase_pb2.NounPhrases
    preposition: str
    target: _noun_phrase_pb2.NounPhrases
    outputs: _containers.RepeatedCompositeFieldContainer[_noun_phrase_pb2.NounPhrases]
    is_read_only: bool
    proper_nouns: _containers.RepeatedScalarFieldContainer[str]

    def __init__(self, english: _Optional[str]=..., verbs: _Optional[_Iterable[str]]=..., object: _Optional[_Union[_noun_phrase_pb2.NounPhrases, _Mapping]]=..., preposition: _Optional[str]=..., target: _Optional[_Union[_noun_phrase_pb2.NounPhrases, _Mapping]]=..., outputs: _Optional[_Iterable[_Union[_noun_phrase_pb2.NounPhrases, _Mapping]]]=..., is_read_only: bool=..., proper_nouns: _Optional[_Iterable[str]]=...) -> None:
        ...


from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Optional as _Optional
DESCRIPTOR: _descriptor.FileDescriptor

class EnvironmentInformationResponse(_message.Message):
    __slots__ = ('version', 'runtime_name', 'runtime_version', 'bci_protocol_version', 'api_version', 'path')
    VERSION_FIELD_NUMBER: _ClassVar[int]
    RUNTIME_NAME_FIELD_NUMBER: _ClassVar[int]
    RUNTIME_VERSION_FIELD_NUMBER: _ClassVar[int]
    BCI_PROTOCOL_VERSION_FIELD_NUMBER: _ClassVar[int]
    API_VERSION_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    version: str
    runtime_name: str
    runtime_version: str
    bci_protocol_version: str
    api_version: str
    path: _containers.RepeatedScalarFieldContainer[str]

    def __init__(self, version: _Optional[str]=..., runtime_name: _Optional[str]=..., runtime_version: _Optional[str]=..., bci_protocol_version: _Optional[str]=..., api_version: _Optional[str]=..., path: _Optional[_Iterable[str]]=...) -> None:
        ...